#!/usr/bin/env python3
#
# Tests the Compiler detection class.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import unittest

import myokit


class CompilerDetectionTest(unittest.TestCase):
    """
    Tests the compiler detection.
    """
    def test_compiler(self):
        # Test the compiler detection.
        self.assertIsInstance(myokit.Compiler.info(), str)


if __name__ == '__main__':
    unittest.main()



#
# Imports a channel model from a ChannelML file.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import io
import os
import textwrap
import warnings
import xml.dom
import xml.dom.minidom

import myokit
from myokit import formats
from myokit import Name, Number, Minus, Multiply, Divide, Power


def dom_child(node):
    """
    Returns the first child element of the given DOM node.

    Returns ``None`` if no such node is found.
    """
    enode = xml.dom.Node.ELEMENT_NODE
    e = node.firstChild
    while e is not None:
        if e.nodeType == enode:
            return e
        e = e.nextSibling
    return None


def dom_next(node, selector=False):
    """
    Returns the next sibling element after the given DOM node.

    If the optional selector is given it searches for an element of a
    particular type.

    Returns ``None`` if no such node is found.
    """
    enode = xml.dom.Node.ELEMENT_NODE
    e = node.nextSibling
    while e is not None:
        if e.nodeType == enode:
            return e
        e = e.nextSibling
    return None


class ChannelMLError(myokit.ImportError):
    """
    Raised if a fatal error occurs when importing ChannelML.
    """


class ChannelMLImporter(formats.Importer):
    """
    This :class:`Importer <myokit.formats.Importer>` imports model definitions
    from the ChannelML format.
    """
    def __init__(self, verbose=False):
        super().__init__()
        self.generated_name_index = 0

    def component(self, path, model):
        """
        Imports a channel component from the ChannelML file at ``path`` and
        adds it to the given model.

        The model **must** contain a variable labelled "membrane_potential".

        The created :class:`myokit.Component` is returned.
        """
        return self._parse(path, model)

    def model(self, path):
        """
        Imports a ChannelML file from ``path`` and returns the channel as a
        :class:`myokit.Model`.
        """
        # Create model
        model = myokit.Model('channelml')
        # Add time variable
        c = model.add_component('engine')
        v = c.add_variable('time')
        v.set_rhs(0)
        v.set_binding('time')
        # Add membrane potential variable
        c = model.add_component('membrane')
        v = c.add_variable('v')
        v.set_rhs('-80')
        v.set_label('membrane_potential')
        # Add the channel component
        self._parse(path, model)
        # Return
        return model

    def _parse(self, path, model):
        """
        Parses a ChannelML channel and adds it to the given model.

        Returns the new :class:`myokit.Component`.
        """
        # Check model: get membrane potential varialbe
        vvar = model.label('membrane_potential')
        if vvar is None:
            raise ChannelMLError(
                'No variable labelled "membrane_potential" was found. This is'
                ' required when adding ChannelML channels to existing models.')

        # Parse XML
        path = os.path.abspath(os.path.expanduser(path))
        dom = xml.dom.minidom.parse(path)

        # Get channelml tag
        root = dom.getElementsByTagName('channelml')
        try:
            root = root[0]
        except IndexError:
            raise ChannelMLError(
                'Unknown root element in xml document. Expecting a tag of type'
                ' <channelml>.')

        # Extract meta data
        meta = self._rip_meta(root)

        # Get channeltype tag
        root = root.getElementsByTagName('channel_type')
        try:
            root = root[0]
        except IndexError:
            raise ChannelMLError(
                'No <channel_type> element found inside <channelml> element.'
                ' Import of <synapse_type> and <ion_concentration> is not'
                ' supported.')

        # Add channel component
        name = self._sanitize_name(root.getAttribute('name'))
        if name in model:
            name_root = name
            i = 2
            while name in model:
                name = name_root + '_' + str(i)
                i += 1
        component = model.add_component(name)

        # Add alias to membrane potential
        component.add_alias('v', vvar)

        # Add meta-data
        component.meta['desc'] = meta

        # Find current-voltage relation
        cvr = root.getElementsByTagName('current_voltage_relation')
        if len(cvr) < 1:
            raise ChannelMLError(
                'Channel model must contain a current voltage relation.')
        elif len(cvr) > 1:
            warnings.warn(
                'Multiple current voltage relations found, ignoring all but'
                ' first.')
        cvr = cvr[0]

        # Check for q10
        try:
            q10 = cvr.getElementsByTagName('q10_settings')[0]
            component.meta['experimental_temperature'] = str(
                q10.getAttribute('experimental_temp'))
        except IndexError:
            pass

        # Add reversal potential
        E = 0
        if cvr.hasAttribute('default_erev'):
            E = float(cvr.getAttribute('default_erev'))
        evar = component.add_variable('E')
        evar.meta['desc'] = 'Reversal potential'
        evar.set_rhs(E)

        # Get maximum conductance
        gmax = 1.0
        if cvr.hasAttribute('default_gmax'):
            gmax = float(cvr.getAttribute('default_gmax'))
        gmaxvar = component.add_variable('gmax')
        gmaxvar.set_rhs(gmax)
        gmaxvar.meta['desc'] = 'Maximum conductance'

        # Add gates
        gvars = []
        for gate in cvr.getElementsByTagName('gate'):
            gname = self._sanitize_name(gate.getAttribute('name'))
            gvar = component.add_variable(gname)
            gvar.promote(0)
            cstate = gate.getElementsByTagName('closed_state')
            cstate = cstate[0].getAttribute('id')
            ostate = gate.getElementsByTagName('open_state')
            ostate = ostate[0].getAttribute('id')

            # Transitions
            trans = gate.getElementsByTagName('transition')
            if len(trans) > 0:
                # Use "transitions" definition
                if len(trans) != 2:
                    raise ChannelMLError(
                        'Expecting exactly 2 transitions for gate <' + gname
                        + '>.')

                # Get closed-to-open state
                tco = None
                for t in trans:
                    if t.getAttribute('to') == ostate and \
                            t.getAttribute('from') == cstate:
                        tco = t
                        break
                if tco is None:
                    raise ChannelMLError(
                        'Unable to find closed-to-open transition for gate <'
                        + gname + '>')

                # Get open-to-closed state
                toc = None
                for t in trans:
                    if t.getAttribute('to') == cstate and \
                            t.getAttribute('from') == ostate:
                        toc = t
                        break
                if toc is None:
                    raise ChannelMLError(
                        'Unable to find open-to-closed transition for gate <'
                        + gname + '>')

                # Add closed-to-open transition
                tname = self._sanitize_name(tco.getAttribute('name'))
                tcovar = gvar.add_variable(tname)
                expr = str(tco.getAttribute('expr'))
                try:
                    tcovar.set_rhs(self._parse_expression(expr, tcovar))
                except myokit.ParseError as e:
                    warnings.warn(
                        'Error parsing expression for closed-to-open'
                        ' transition in gate <' + gname + '>: '
                        + myokit.format_parse_error(e))
                    tcovar.meta['expression'] = str(expr)

                # Add open-to-closed transition
                tname = self._sanitize_name(toc.getAttribute('name'))
                tocvar = gvar.add_variable(tname)
                expr = str(toc.getAttribute('expr'))
                try:
                    tocvar.set_rhs(self._parse_expression(expr, tocvar))
                except myokit.ParseError as e:
                    warnings.warn(
                        'Error parsing expression for open-to-closed'
                        ' transition in gate <' + gname + '>: '
                        + myokit.format_parse_error(e))
                    tocvar.meta['expression'] = str(expr)

                # Write equation for gate
                gvar.set_rhs(Minus(
                    Multiply(Name(tcovar), Minus(Number(1), Name(gvar))),
                    Multiply(Name(tocvar), Name(gvar))))

            else:
                # Use "steady-state & time_course" definition
                ss = gate.getElementsByTagName('steady_state')
                tc = gate.getElementsByTagName('time_course')
                if len(ss) < 1 or len(tc) < 1:
                    raise ChannelMLError(
                        'Unable to find transitions or steady state and'
                        ' time course for gate <' + gname + '>.')
                ss = ss[0]
                tc = tc[0]

                # Add steady-state variable
                ssname = self._sanitize_name(ss.getAttribute('name'))
                ssvar = gvar.add_variable(ssname)
                expr = str(ss.getAttribute('expr'))
                try:
                    ssvar.set_rhs(self._parse_expression(expr, ssvar))
                except myokit.ParseError as e:
                    warnings.warn(
                        'Error parsing expression for steady state in gate <'
                        + gname + '>: ' + myokit.format_parse_error(e))
                    ssvar.meta['expression'] = str(expr)

                # Add time course variable
                tcname = self._sanitize_name(tc.getAttribute('name'))
                tcvar = gvar.add_variable(tcname)
                expr = str(tc.getAttribute('expr'))
                try:
                    tcvar.set_rhs(self._parse_expression(expr, tcvar))
                except myokit.ParseError as e:
                    warnings.warn(
                        'Error parsing expression for time course in gate <'
                        + gname + '>: ' + myokit.format_parse_error(e))
                    tcvar.meta['expression'] = str(expr)

                # Write expression for gate
                gvar.set_rhs(
                    Divide(Minus(Name(ssvar), Name(gvar)), Name(tcvar)))

            power = int(gate.getAttribute('instances'))
            if power > 1:
                gvars.append(Power(Name(gvar), Number(power)))
            else:
                gvars.append(Name(gvar))

        if len(gvars) < 1:
            raise ChannelMLError(
                'Current voltage relation requires at least one gate.')

        # Add current variable
        ivar = component.add_variable('I')
        ivar.meta['desc'] = 'Current'
        expr = Name(gmaxvar)
        while gvars:
            expr = Multiply(expr, gvars.pop())
        expr = Multiply(expr, Minus(Name(vvar), Name(evar)))
        ivar.set_rhs(expr)

        # Done, return component
        return component

    def _parse_expression(self, s, var=None):
        """
        Attempts to read an expression from the given string ``s`` and parse it
        using the context of variable ``var`` to resolve any references.
        """
        s = str(s)
        # Pre-process some basic html entities
        entities = {
            '&eq;': '==',
            '&ne;': '!=',
            '&neq;': '!=',  # Not a html entity!
            '&lt;': '<',
            '&gt;': '>',
            '&le;': '<=',
            '&ge;': '>=',
        }
        s = s.replace('&amp;', '&')
        for k, v in entities.items():
            s = s.replace(k, v)
        # Attempt to handle single a?b:c construct
        if '?' in s:
            c, v = s.split('?', 1)
            if ':' in v:
                v = v.split(':', 1)
            else:
                v = (v, '0')
            return myokit.If(
                myokit.parse_expression(c, var),
                myokit.parse_expression(v[0], var),
                myokit.parse_expression(v[1], var))
        return myokit.parse_expression(s, var)

    def _rip_meta(self, root):
        """
        Coarsely extracts any meta data from the model and returns it as a text
        file.
        """
        if not root.hasAttribute('xmlns:meta'):
            return None
        ns = root.getAttribute('xmlns:meta')
        b = io.StringIO()

        def scan(parent):
            kid = dom_child(parent)
            while kid is not None:
                t = None
                if kid.namespaceURI == ns:
                    t = self._flatten(kid)
                if t:
                    b.write(t)
                    b.write('\n')
                else:
                    scan(kid)
                kid = dom_next(kid)
        scan(root)
        return b.getvalue().strip()

    def _flatten(self, node):
        """
        Reduces a node's contents to flat text content and returns it.
        """
        def text(node, buff=None):
            if buff is None:
                buff = []
            if node.nodeType == node.TEXT_NODE:
                t = node.nodeValue
                t = t.replace('\n', ' ')
                t = t.replace('\r', ' ')
                t = t.replace('\f', ' ')
                t = t.strip()
                if t != '':
                    buff.append(t)
            else:
                for kid in node.childNodes:
                    text(kid, buff)
            return buff
        return textwrap.fill(
            str('\n'.join(text(node))), 70, replace_whitespace=False)

    def _sanitize_name(self, name):
        """
        Tests if a name is a valid myokit name. Adapts it if it isn't.
        """
        name = str(name).strip()
        name = name.replace('.', '_')
        name = name.replace(' ', '_')
        name = name.replace('-', '_')
        try:
            myokit.check_name(name)
        except myokit.InvalidNameError as e:
            warnings.warn('Invalid name: ' + str(e))
            self.generated_name_index += 1
            name = 'generated_name_' + str(self.generated_name_index)
        return name

    def supports_component(self):
        return True

    def supports_model(self):
        return True




#!/usr/bin/env python3
#
# Tests meta data functions in myokit.Model
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import unittest

import myokit


class MetaTest(unittest.TestCase):
    """
    Tests meta data functions in myokit.Model.
    """

    def test_basic(self):
        # Test the basic meta-data functionality

        # Test assignment with simple and namespaced names
        m = myokit.Model()
        m.meta['bert'] = 'ernie'
        m.meta['bert:ernie'] = 'banaan'

        def f():
            m.meta['bert.ernie'] = 'banaan'
        self.assertRaises(myokit.InvalidMetaDataNameError, f)

        def f():
            m.meta['bert '] = 'banaan'
        self.assertRaises(myokit.InvalidMetaDataNameError, f)

        # Test retrieval
        self.assertEqual(m.meta['bert'], 'ernie')
        self.assertEqual(m.meta['bert:ernie'], 'banaan')

        def f():
            return m.meta['bert.ernie']
        self.assertRaises(myokit.InvalidMetaDataNameError, f)

        # Test overwriting
        m.meta['bert'] = 'verrekijker'
        self.assertEqual(m.meta['bert'], 'verrekijker')

        # Test deletion
        del m.meta['bert']
        self.assertRaises(KeyError, lambda: m.meta['bert'])

    def test_parser(self):
        # Test basic meta-data parsing

        # Parse simple model without meta data
        lines = [
            '[[model]]',
            'a.x = 0',
            '[a]',
            't = 0 bind time',
            'dot(x) = 4',
        ]
        m = myokit.parse_model('\n'.join(lines))
        self.assertTrue(m.is_valid())

        # Test simple meta data
        def model(key, value):
            x = list(lines)
            x.insert(1, key + ':' + value)
            m = myokit.parse_model('\n'.join(x))
            self.assertTrue(m.is_valid())
            self.assertEqual(m.meta[key], value)

        def component(key, value):
            m = myokit.parse_model('\n'.join(
                lines + [key + ':' + value]))
            self.assertTrue(m.is_valid())
            self.assertEqual(m.get('a').meta[key], value)

        def variable(key, value):
            m = myokit.parse_model('\n'.join(
                lines + ['    ' + key + ':' + value]))
            self.assertTrue(m.is_valid())
            self.assertEqual(m.get('a.x').meta[key], value)

        def test(key, value):
            model(key, value)
            component(key, value)
            variable(key, value)

        test('vic', 'bob')
        # Test empty property
        test('vic', '')
        # Test namespaced keys
        test('vic:bob', 'uvavu')
        test('vic:bob:eranu', 'uvavu')
        # Test invalid keys
        self.assertRaises(myokit.ParseError, test, 'vic.bob', 'uvavu')


if __name__ == '__main__':
    unittest.main()



#
# Exports to Ansi C, using the CVODE libraries for integration
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os

import myokit.formats


class AnsiCExporter(myokit.formats.TemplatedRunnableExporter):
    """
    This :class:`Exporter <myokit.formats.Exporter>` generates a runnable ansic
    C model implementation and integrator. The integration is based on the
    Sundials CVODE library, which is required to run the program.

    Both the model definition and pacing protocol are exported and the file
    is set up ready to run a simulation. No post-processing is included.

    Provides the following external variables:

    ``time``
        The current simulation time
    ``pace``
        The current value of the pacing system, implemented using the given
        protocol.
    """
    def post_export_info(self):
        return '\n'.join((
            'To compile in gcc, use::',
            '',
            '    gcc -Wall -lm -lsundials_cvode sim.c -o sim',    # noqa
            '',
            'Example plot script::',
            '',
            '    import matplotlib.pyplot as plt',
            '    with open("v.txt", "r") as f:',
            '      T, V = [], []',
            '      for line in f:',
            '        t, v = [float(x) for x in line.split()]',
            '        T.append(t)',
            '        V.append(v)',
            '    plt.figure()',
            '    plt.plot(T, V)',
            '    plt.show()',
        ))

    def _dir(self, root):
        return os.path.join(root, 'ansic', 'template')

    def _dict(self):
        return {'sim.c': 'sim.c'}

    def _vars(self, model, protocol):
        return {'model': model, 'protocol': protocol}


class AnsiCEulerExporter(myokit.formats.TemplatedRunnableExporter):
    """
    This :class:`Exporter <myokit.formats.Exporter>` generates an ansic C
    implementation using a simple explicit forward-Euler scheme.

    Both the model definition and pacing protocol are exported and the file
    is set up ready to run a simulation. No post-processing is included.

    Provides the following external variables:

    ``time``
        The simulation time
    ``pace``
        The value of the pacing variable, implemented using the given protocol.

    No labeled variables are required.
    """
    def post_export_info(self):
        return '\n'.join((
            'To compile using gcc::',
            '',
            '    gcc -Wall -lm euler.c -o euler',
            '',
            'Example plot script:',
            '',
            '    import matplotlib.pyplot as plt',
            '    with open("v.txt", "r") as f:',
            '        T, V = [], []',
            '        for line in f:',
            '          t, v = [float(x) for x in line.split()]',
            '          T.append(t)',
            '          V.append(v)',
            '    plt.figure()',
            '    plt.plot(T, V)',
            '    plt.show()',
        ))

    def _dir(self, root):
        return os.path.join(root, 'ansic', 'template')

    def _dict(self):
        return {'euler.c': 'euler.c'}

    def _vars(self, model, protocol):
        return {'model': model, 'protocol': protocol}


class AnsiCCableExporter(myokit.formats.TemplatedRunnableExporter):
    """
    This :class:`Exporter <myokit.formats.Exporter>` generates a 1d cable
    simulation using a simple forward-Euler scheme in ansi-C.

    Both the model definition and pacing protocol are exported and the file
    is set up ready to run a simulation. No post-processing is included.

    Provides the following external variables:

    ``time``
        The simulation time
    ``pace``
        The value of the pacing variable, implemented using the given protocol.
    ``diffusion_current``
        The current flowing from each cell to its neighbors. This will be
        positive if the cell is acting as a source, negative when it's acting
        as a sink.

    Requires the following labels to be set in the model:

    ``membrane_potential``
        The membrane potential.

    Variables are linked using ``diffusion_current``, this is calculated from
    the membrane potentials as:

        i = g * ((V - V_next) - (V_last - V))

    At the boundaries, V is substituted for V_last or V_next.
    """
    def post_export_info(self):
        return '\n'.join((
            'To compile using gcc::',
            '',
            '    gcc -Wall -lm cable.c -o cable',
            '',
            'Example plot script::',
            '',
            '    import myokit',
            '    import numpy as np',
            '    import matplotlib.pyplot as plt',
            '    from mpl_toolkits.mplot3d import axes3d',
            '    d = myokit.load_csv("data.txt")',
            '    n = 50 # Assuming 50 cells',
            '    f = plt.figure()',
            '    x = f.gca(projection="3d")',
            '    z = np.ones(len(d["time"]))',
            '    for i in range(0, n):',
            '        x.plot(d["time"], z*i, d.get(str(i) + "_V"))',
            '    plt.show()',
        ))

    def _dir(self, root):
        return os.path.join(root, 'ansic', 'template')

    def _dict(self):
        return {'cable.c': 'cable.c'}

    def _vars(self, model, protocol):
        return {'model': model, 'protocol': protocol}



#
# OpenCL expression writer
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import myokit

from myokit.formats.python import PythonExpressionWriter


class OpenCLExpressionWriter(PythonExpressionWriter):
    """
    This :class:`ExpressionWriter <myokit.formats.ExpressionWriter>` translates
    Myokit :class:`expressions <myokit.Expression>` to OpenCL syntax.
    """
    def __init__(self, precision=myokit.SINGLE_PRECISION, native_math=True):
        super().__init__()
        self._function_prefix = ''
        self._sp = (precision == myokit.SINGLE_PRECISION)
        self._nm = bool(native_math)

    def _exc(self, e):
        """Returns ``ex(e)`` if ``e`` is a Condition, else ``ex(e != 0)``."""
        if isinstance(e, myokit.Condition):
            return self.ex(e)
        return self.ex(myokit.NotEqual(e, myokit.Number(0)))

    def _ex_infix_condition(self, e, op):
        """Handles ex() for infix condition operators (==, !=, > etc.)."""
        c1 = isinstance(e[0], myokit.Condition)
        c2 = isinstance(e[1], myokit.Condition)
        if (c1 and c2) or not (c1 or c2):
            a = self.ex(e[0])
            b = self.ex(e[1])
        else:
            a = self._exc(e[0])
            b = self._exc(e[1])
        return '(' + a + ' ' + op + ' ' + b + ')'

    def _ex_infix_logical(self, e, op):
        """Handles ex() for infix logical operators."""
        return '(' + self._exc(e[0]) + ' ' + op + ' ' + self._exc(e[1]) + ')'

    #def _ex_name(self, e):
    #def _ex_derivative(self, e):

    def _ex_number(self, e):
        return myokit.float.str(e) + 'f' if self._sp else myokit.float.str(e)

    #def _ex_prefix_plus(self, e):
    #def _ex_prefix_minus(self, e):
    #def _ex_plus(self, e):
    #def _ex_minus(self, e):
    #def _ex_multiply(self, e):

    def _ex_divide(self, e):
        # Native divide seemed to cause some issues
        #if self._nm:
        #    return 'native_divide(' + self.ex(e[0]) +', '+ self.ex(e[1]) + ')'
        #else:
        return self._ex_infix(e, '/')

    def _ex_quotient(self, e):
        # Note that this _must_ round towards minus infinity.
        # See myokit.Quotient.
        # Assuming it follows C and so we need a custom implementation.
        return self.ex(myokit.Floor(myokit.Divide(e[0], e[1])))

    def _ex_remainder(self, e):
        # Note that this _must_ use the same round-to-neg-inf convention as
        # myokit.Quotient.
        # Assuming it follows C and so we need a custom implementation.
        return self.ex(myokit.Minus(
            e[0], myokit.Multiply(e[1], myokit.Quotient(e[0], e[1]))))

    def _ex_power(self, e):
        if e[1] == myokit.Number(2):
            if e.bracket(e[0]):
                return '((' + self.ex(e[0]) + ') * (' + self.ex(e[0]) + '))'
            else:
                return '(' + self.ex(e[0]) + ' * ' + self.ex(e[0]) + ')'
        else:
            return 'pow(' + self.ex(e[0]) + ', ' + self.ex(e[1]) + ')'

    def _ex_sqrt(self, e):
        f = 'native_sqrt' if self._nm else 'sqrt'
        return self._ex_function(e, f)

    def _ex_sin(self, e):
        f = 'native_sin' if self._nm else 'sin'
        return self._ex_function(e, f)

    def _ex_cos(self, e):
        f = 'native_cos' if self._nm else 'cos'
        return self._ex_function(e, f)

    def _ex_tan(self, e):
        f = 'native_tan' if self._nm else 'tan'
        return self._ex_function(e, f)

    def _ex_asin(self, e):
        return self._ex_function(e, 'asin')

    def _ex_acos(self, e):
        return self._ex_function(e, 'acos')

    def _ex_atan(self, e):
        return self._ex_function(e, 'atan')

    def _ex_exp(self, e):
        f = 'native_exp' if self._nm else 'exp'
        return self._ex_function(e, f)

    def _ex_log(self, e):
        f = 'native_log' if self._nm else 'log'
        if len(e) == 1:
            return self._ex_function(e, f)
        return '(' + f + '(' + self.ex(e[0]) + ') / ' + f + '(' \
            + self.ex(e[1]) + '))'

    def _ex_log10(self, e):
        f = 'native_log10' if self._nm else 'log10'
        return self._ex_function(e, f)

    def _ex_floor(self, e):
        return self._ex_function(e, 'floor')

    def _ex_ceil(self, e):
        return self._ex_function(e, 'ceil')

    def _ex_abs(self, e):
        return self._ex_function(e, 'fabs')

    def _ex_not(self, e):
        return '!(' + self._exc(e[0]) + ')'

    #def _ex_equal(self, e):
    #def _ex_not_equal(self, e):
    #def _ex_more(self, e):
    #def _ex_less(self, e):
    #def _ex_more_equal(self, e):
    #def _ex_less_equal(self, e):

    def _ex_and(self, e):
        return self._ex_infix_logical(e, '&&')

    def _ex_or(self, e):
        return self._ex_infix_logical(e, '||')

    def _ex_if(self, e):
        return '(%s ? %s : %s)' % (
            self._exc(e._i), self.ex(e._t), self.ex(e._e))

    def _ex_piecewise(self, e):
        s = []
        n = len(e._i)
        for i in range(0, n):
            s.append('(')
            s.append(self._exc(e._i[i]))
            s.append(' ? ')
            s.append(self.ex(e._e[i]))
            s.append(' : ')
        s.append(self.ex(e._e[n]))
        s.append(')' * n)
        return ''.join(s)




#!/usr/bin/env python3
#
# Tests the myokit.lib.multi module.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import unittest

import myokit

from myokit.tests import DIR_DATA, WarningCollector

# Path to multi-model testing files
DIR_MULTI = os.path.join(DIR_DATA, 'multi')


class LibMultiTest(unittest.TestCase):
    """
    Tests the myokit.lib.multi module.
    """

    def test_iterdir(self):
        # Test the iterdir() method that iterators over model, protocol tuples.

        with WarningCollector():
            import myokit.lib.multi as multi

        # Get all found tuples (model, protocol)
        tuples = [x for x in multi.iterdir(DIR_MULTI)]
        self.assertEqual(len(tuples), 2)

        # Should contain (ordered by filename)
        #  (Beeler (no name), None)
        #  (Lr1991, protocol)
        #
        self.assertEqual(type(tuples[0][0]), myokit.Model)
        self.assertEqual(tuples[0][0].name(), 'beeler-no-name')
        self.assertIsNone(tuples[0][1], None)
        self.assertEqual(type(tuples[1][0]), myokit.Model)
        self.assertEqual(tuples[1][0].name(), 'Luo-Rudy model (1991)')
        self.assertEqual(type(tuples[1][1]), myokit.Protocol)

        # Test without name setting
        tuples = [x for x in multi.iterdir(DIR_MULTI, False)]
        self.assertEqual(len(tuples), 2)
        self.assertEqual(type(tuples[0][0]), myokit.Model)
        self.assertIsNone(tuples[0][0].name())
        self.assertIsNone(tuples[0][1], None)
        self.assertEqual(type(tuples[1][0]), myokit.Model)
        self.assertEqual(tuples[1][0].name(), 'Luo-Rudy model (1991)')
        self.assertEqual(type(tuples[1][1]), myokit.Protocol)

        # Path must be a directory
        path = os.path.join(DIR_MULTI, 'lr-1991.mmt')
        i = multi.iterdir(path)
        self.assertRaisesRegex(ValueError, 'not a directory', next, i)

    def test_scandir(self):
        # Test the scandir() method that returns a models list and a protocols
        # list.

        with WarningCollector():
            import myokit.lib.multi as multi

        # Get list of models and protocols
        models, protocols = multi.scandir(DIR_MULTI)
        self.assertEqual(len(models), len(protocols), 2)

        # Should contain (ordered by model name)
        #  (Lr1991, protocol) --> Upper case goes first
        #  (Beeler (no name), None)
        #
        self.assertEqual(type(models[0]), myokit.Model)
        self.assertEqual(type(models[1]), myokit.Model)
        self.assertEqual(models[0].name(), 'Luo-Rudy model (1991)')
        self.assertEqual(models[1].name(), 'beeler-no-name')
        self.assertEqual(type(protocols[0]), myokit.Protocol)
        self.assertIsNone(protocols[1])

    def test_time(self):
        # Test the time() method that returns the time variable.

        with WarningCollector():
            import myokit.lib.multi as multi

        m = myokit.Model()
        c = m.add_component('c')
        x = c.add_variable('x')
        x.set_rhs(0)

        with WarningCollector() as w:
            self.assertRaises(myokit.IncompatibleModelError, multi.time, m)
            x.set_binding('time')
            self.assertEqual(x, multi.time(m))
        self.assertIn('deprecated', w.text())

    def test_label(self):
        # Test the label() method that returns a labelled variable.

        with WarningCollector():
            import myokit.lib.multi as multi

        m = myokit.Model()
        c = m.add_component('c')
        x = c.add_variable('x')
        x.set_rhs(0)

        with WarningCollector() as w:
            self.assertRaises(
                myokit.IncompatibleModelError, multi.label, m, 'x')
            x.set_label('x')
            self.assertEqual(x, multi.label(m, 'x'))
        self.assertIn('deprecated', w.text())

    def test_binding(self):
        # Test the binding() method that returns a bound variable.

        with WarningCollector():
            import myokit.lib.multi as multi

        m = myokit.Model()
        c = m.add_component('c')
        x = c.add_variable('x')
        x.set_rhs(0)

        with WarningCollector() as w:
            self.assertRaises(
                myokit.IncompatibleModelError, multi.binding, m, 'x')
            x.set_binding('x')
            self.assertEqual(x, multi.binding(m, 'x'))
        self.assertIn('deprecated', w.text())

    def test_unit(self):
        # Test the unit() method that returns a unit conversion factor.

        with WarningCollector():
            import myokit.lib.multi as multi

        m = myokit.Model()
        c = m.add_component('c')
        x = c.add_variable('x')
        x.set_rhs(0)
        x.set_unit(myokit.parse_unit('mV'))

        self.assertRaises(
            myokit.IncompatibleModelError, multi.unit, x,
            myokit.parse_unit('kg'))
        self.assertEqual(multi.unit(x, myokit.parse_unit('V')), 0.001)
        self.assertEqual(multi.unit(x, myokit.parse_unit('mV')), 1)
        self.assertEqual(multi.unit(x, myokit.parse_unit('uV')), 1000)


if __name__ == '__main__':
    unittest.main()



#
# Methods for guessing the meaning of model variables (and manipulating models
# based on those guesses).
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import collections

import myokit


def _compatible_units(unit, options):
    """ Checks if ``unit`` is compatible with one of the given ``options``. """
    if unit is None:
        return False
    for option in options:
        if myokit.Unit.can_convert(unit, option):
            return True
    return False


def _deep_deps(variable):
    """
    Finds all variables that ``variable`` depends on, and returns a map
    ``{var: distance to var}``.
    """
    distances = {variable: 0}
    queue = collections.deque([variable])
    while queue:
        root = queue.popleft()
        dist = distances[root] + 1
        for v in root.refs_to():
            if v not in distances:
                distances[v] = dist
                queue.append(v)
    del distances[variable]

    return distances


def _distance_to_bound(variable):
    """
    Finds all variables that depend on a bound ``variable``, but are otherwise
    constant, and returns a map ``{var: distance to var}``.
    """
    # 1. Remove binding
    # 2. Find distances, scanning only constants dependent on `variable`
    # 3. Reinstate binding

    # Remove binding
    binding = variable.binding()
    if binding is None:
        raise ValueError('Argument `variable` must be a bound variable.')
    variable.set_binding(None)

    # Find distances
    try:
        distances = {variable: 0}
        queue = collections.deque([variable])
        while queue:
            root = queue.popleft()
            dist = 1 + distances[root]
            for v in root.refs_by():
                if v.is_constant() and v not in distances:
                    distances[v] = dist
                    queue.append(v)
        del distances[variable]
    finally:
        # Reinstate binding
        variable.set_binding(binding)

    return distances


def _find_conditional(expression):
    """
    Performs a breadth-first search on the given ``expression``, returning the
    first conditional expression it finds (or ``None`` if no conditionals are
    encountered).
    """
    conds = (myokit.If, myokit.Piecewise)
    if isinstance(expression, conds):
        return expression

    queue = collections.deque([expression])
    while queue:
        root = queue.popleft()
        for e in iter(root):
            if isinstance(e, conds):
                return e
            queue.append(e)

    return None


def _remove_nested(variable):
    """
    Removes any nested variables of the given variable.
    """
    kids = [k for k in variable]
    for k in kids:
        k.set_rhs(0)
    for k in kids:
        variable.remove_variable(k, recursive=True)


def add_embedded_protocol(model, protocol, add_oxmeta_annotations=True):
    """
    Attempts to convert a :class:`myokit.Protocol` to a (discontinuous)
    expression and embed it in the given :class`myokit.Model`.

    Returns ``True`` if the model was succesfully updated.

    This method is designed for protocols that contain a single, indefinitely
    recurring event (e.g. cell pacing protocols), and doesn't handle other
    forms of protocol.

    If ``add_oxmeta_annotations`` is set to ``True``, the method will add
    the relevant ``oxmeta`` annotations to any newly created variables
    """

    # Get time variable
    time = model.time()
    if time is None:
        return False

    # Get pacing variable
    v_pace = model.binding('pace')
    if v_pace is None:
        # Nothing bound to pace: Might be able to add protocol, but there'd be
        # no point to it!
        return False

    # Check protocol has only a single event
    if len(protocol) != 1:
        return False

    # Check it's an indefinitely recurring periodic event
    event = protocol.head()
    if event.period() == 0 or event.multiplier() != 0:
        return False

    # Check for existing stimulus variables
    stim_info = stimulus_current_info(model)
    v_current = stim_info['current']
    v_amplitude = stim_info['amplitude']
    e_amplitude = stim_info['amplitude_expression']
    v_offset = stim_info['offset']
    v_period = stim_info['period']
    v_duration = stim_info['duration']

    # Don't add a stimulus if there's already periodic stimulus variables about
    if v_offset is not None or v_period is not None or v_duration is not None:
        return False
    del v_offset, v_period, v_duration

    # Determine variable to update: pace or i_stim
    if v_current is None or (v_amplitude is None and e_amplitude is None):
        # Unable to set current as (...) * amplitude? Then update v_pace
        updating_current = False

        # VarOwner to add variables to
        parent = v_pace

    else:
        # Able to set current as (...) * amplitude? Then update v_current
        updating_current = True

        # VarOwner to add variables to
        parent = v_current

    # Start modifying model

    # Add new child variables with stimulus properties
    v_offset = parent.add_variable_allow_renaming('offset')
    v_offset.set_unit(time.unit())
    v_offset.set_rhs(myokit.Number(event.start(), time.unit()))

    v_period = parent.add_variable_allow_renaming('period')
    v_period.set_unit(time.unit())
    v_period.set_rhs(myokit.Number(event.period(), time.unit()))

    v_duration = parent.add_variable_allow_renaming('duration')
    v_duration.set_unit(time.unit())
    v_duration.set_rhs(myokit.Number(event.duration(), time.unit()))

    # Add oxmeta annotations for web lab
    if updating_current and add_oxmeta_annotations:
        v_offset.meta['oxmeta'] = 'membrane_stimulus_current_offset'
        v_period.meta['oxmeta'] = 'membrane_stimulus_current_period'
        v_duration.meta['oxmeta'] = 'membrane_stimulus_current_duration'
        if v_amplitude is not None and 'oxmeta' not in v_amplitude.meta:
            v_amplitude.meta['oxmeta'] = 'membrane_stimulus_current_amplitude'

    # Create expression for pacing signal
    pace_term = myokit.If(
        myokit.Less(
            myokit.Remainder(
                myokit.Minus(myokit.Name(time), myokit.Name(v_offset)),
                myokit.Name(v_period)
            ),
            myokit.Name(v_duration)
        ),
        myokit.Number(event.level(), v_pace.unit()),
        myokit.Number(0, v_pace.unit())
    )

    if updating_current:

        # Update current variable
        if e_amplitude is None:
            e_amplitude = myokit.Name(v_amplitude)
        v_current.set_rhs(myokit.Multiply(pace_term, e_amplitude))

        # Unbind pacing variable and remove if unused
        v_pace.set_binding(None)
        v_pace.set_rhs(myokit.Number(0, v_pace.unit()))
        _remove_nested(v_pace)
        if len(list(v_pace.refs_by())) == 0:
            v_pace.parent().remove_variable(v_pace)

    else:

        # Update pacing variable
        v_pace.set_rhs(pace_term)

        # Unbind pacing variable
        v_pace.set_binding(None)

    # It worked!
    return True


def membrane_capacitance(model):
    """
    Guess which model variable (if any) represents the membrane capacitance.

    The following strategy is used:

    1. If a variable is found with the annotation
       ``oxmeta: membrane_capacitance``, this variable is returned.
    2. If not, a list of candidates is drawn up from all constants in the
       model.
    3. 1.5 points are awarded for being referenced by the membrane potential.
    4. A point is awarded for having a unit compatible with ``F``, ``F/m^2``,
       or ``m^2``. A point is subtracted for having a unit other than ``None``
       that's incompatible with any of the above.
    5. A point is awarded for having the name ``C``, ``Cm``, ``C_m``, ``Acap``
       or ``A_cap``, or  having a name including the string 'capacitance'
       (where the checks are made case-insensitively).

    If no suitable candidates are found the method returns ``None``, otherwise
    the highest ranking candidate is returned.
    """
    # Return any variable annotated as `oxmeta: membrane_capacitance`
    for v in model.variables(deep=True):
        if v.meta.get('oxmeta', None) == 'membrane_capacitance':
            return v

    # Gather candidates
    candidates = {v: 0 for v in model.variables(deep=True, const=True)}

    # Points if used by Vm
    vm = membrane_potential(model)
    if vm is not None:
        for ref in vm.refs_to():
            if ref.is_constant():
                candidates[ref] += 1.5

    # Points for units
    def is_compatible(unit1, unit2):
        try:
            myokit.Unit.conversion_factor(unit1, unit2)
            return True
        except myokit.IncompatibleUnitError:
            return False

    cap1 = myokit.units.farad
    cap2 = myokit.units.farad / myokit.units.m**2
    cap3 = myokit.units.m**2
    for v in candidates:
        unit = v.unit()
        if unit is None:
            continue
        elif is_compatible(unit, cap1) or is_compatible(unit, cap2):
            candidates[v] += 1
        elif is_compatible(unit, cap3):
            candidates[v] += 1
        else:
            candidates[v] -= 1

    # Points for names
    names = ['c', 'cm', 'c_m', 'acap', 'a_cap']
    for v in candidates:
        name = v.name().lower()
        if name in names or 'capacitance' in name:
            candidates[v] += 1

    # Find best candidate, with at least some credibility
    if candidates:
        ranking = sorted(candidates.keys(), key=lambda v: -candidates[v])
        v = ranking[0]
        if candidates[v] > 0:
            return v

    # Nothing found!
    return None


def membrane_currents(model):
    """
    Gueses which model variables represent ionic currents through the outer
    membrane and returns them.

    This method assumes that all currents follow the same sign convention.

    Currents defined as sums of currents (or multiples of fluxes) are not
    expanded.
    """
    # Get membrane potential
    vm = membrane_potential(model)

    # Get expression that refers to currents (possibly indirectly)
    i_ion = model.label('cellular_current')
    if i_ion is not None:
        e_currents = i_ion.rhs()
    else:
        e_currents = vm.rhs()
    del i_ion

    # Assume that e_currents is an expression such as:
    #  INa + ICaL + IKr + ...
    #  i_ion + i_diff + i_stim
    #  -1/C * (...)
    currents = []
    for term in e_currents.references():
        if term.is_constant():
            continue

        # Get current variable
        current = term.var()
        currents.append(current)

    # Sort and return
    currents.sort(key=lambda x: x.name())
    return currents


def membrane_potential(model):
    """
    Gueses which model variable (if any) represents the membrane potential and
    returns it.

    Three strategies can be attempted.

    First, variables annotated as membrane potential will be returned,
    regardless of name, units, etc. In order of priority this method will
    search for (1) the label ``membrane_potential``, (2) the meta data property
    ``oxmeta: membrane_voltage``.

    If no annotated variables are found, the following strategy is used:

    1. A list of candidates is compiled, consisting of all non-nested
       variables.
    2. Candidates with a unit compatible with voltage get +1 points
    3. Candidates that define a unit (not ``None``) incompatible with voltage
       are discarded
    4. Candidates in a component with a common name for membrane potential
       variables get +1 point. Similarly candidates in a common component for
       membrane potential get +1 point, with an additional +0.5 point if both
       conditions are met.
    5. State variables get +1 point.
    6. The candidate that is used in the most equations gets +0.1 point.
    7. Candidates with a score < 1 are discarded.
    8. The highest scoring candidate is returned (with no particular
       tie-breaking rules).

    If all strategies fail the method returns ``None``.
    """
    # Return any variable labelled as `membrane_potential'
    v = model.label('membrane_potential')
    if v is not None:
        return v

    # Return any variable annotated as `oxmeta: membrane_stimulus_current`
    for v in model.variables(deep=True):
        if v.meta.get('oxmeta', None) == 'membrane_voltage':
            return v

    # Common units for the membrane potential
    common_units = [myokit.units.V]

    # Non-dimensionalized (e.g. Mitchell-Schaeffer) use [1]
    # But this causes too many false positives
    # common_units.append(myokit.units.dimensionless)

    # Common names for the membrane potential
    common_component_names = [
        'membrane',
        'cell',
    ]
    common_names = [
        'v',
        'vm',
        'v_m',
    ]

    # Create a list of candidates, with initial scores based on units
    candidates = {}
    for c in model.components():
        for v in model.variables():  # No nested variables
            # Ignore time variable
            if v.binding() == 'time':
                continue
            unit = v.unit()
            if unit is None:
                candidates[v] = 0
            elif _compatible_units(unit, common_units):
                candidates[v] = 1

    # Easy case: no candidates or a single candidate
    if len(candidates) == 0:
        return None
    elif len(candidates) == 1:
        return candidates.popitem()[0]

    # Award points for names
    for v in candidates:
        cname = v.parent().name().lower() in common_component_names
        vname = v.name().lower() in common_names
        score = int(cname) + int(vname)
        if cname and vname:
            score += 0.5
        if score:
            candidates[v] += score

    # Award point for states
    for v in candidates:
        if v.is_state():
            candidates[v] += 1

    # Award tie-breaking point for most references
    ranking = list(candidates)
    ranking.sort(key=lambda v: -len(list(v.refs_by())))
    candidates[ranking[0]] += 0.1

    # Find (one of) best candidate
    ranking.sort(key=lambda v: -candidates[v])

    # Check candidate has at least 1 full point and return
    v = ranking[0]
    if candidates[v] < 1:
        return None
    return v


def remove_embedded_protocol(model):
    """
    Searches the given model for a hardcoded periodic stimulus current and, if
    one is found, removes it and returns a :class:`myokit.Protocol` instead.

    For this method to work, :meth:`stimulus_current_info` should return at
    least a current variable, a period and duration variable, and either an
    amplitude variable or an amplitude expression.

    Returns a :class:`myokit.Protocol` if succesful, or ``None`` if not.
    """

    # Get stimulus current info from model
    info = stimulus_current_info(model)

    # Must have a stimulus current
    v_current = info['current']
    if v_current is None:
        return None

    # Must have a period and duration
    v_duration = info['duration']
    v_period = info['period']
    if v_period is None or v_duration is None:
        return None
    period = v_period.eval()
    duration = v_duration.eval()

    # Check values are sensible
    if period <= 0 or duration <= 0:
        return None

    # Offset is optional
    v_offset = info['offset']
    if v_offset is None:
        offset = 0
    else:
        offset = v_offset.eval()

    # Must have an amplitude or an amplitude expression
    v_amplitude = info['amplitude']
    e_amplitude = info['amplitude_expression']
    if e_amplitude is None:
        if v_amplitude is None:
            return None
        e_amplitude = myokit.Name(v_amplitude)

    # Start modifying model

    # Get pacing variable
    v_pace = model.binding('pace')
    if v_pace is None:
        v_pace = v_current.parent(
            myokit.Component).add_variable_allow_renaming('pace')
        v_pace.set_unit(myokit.units.dimensionless)
        v_pace.set_rhs(myokit.Number(0, myokit.units.dimensionless))
        v_pace.set_binding('pace')

    # Set new RHS for stimulus variable
    v_current.set_rhs(myokit.Multiply(myokit.Name(v_pace), e_amplitude))

    # Remove stimulus variables, if unused
    for v in (v_period, v_duration, v_offset, v_amplitude):
        if v is not None and len(list(v.refs_by())) == 0:
            v.parent().remove_variable(v, recursive=True)

    # Return periodic protocol
    return myokit.pacing.blocktrain(
        period=period, duration=duration, offset=offset)


def stimulus_current(model):
    """
    Guesses which model variable (if any) represents the stimulus current and
    returns it.

    Three strategies are tried to find the stimulus current.

    Firstly, annotated variables are searched for. If found, these are returned
    without further checking of names, units etc. Supported annotations (in
    order of priority) are (1) the label ``stimulus_current``, (2) the meta
    data property ``oxmeta: membrane_stimulus_current``.

    If no annotated variables are found, the following strategy is followed:

    1. A list of candidate variables that depend (in)directly on ``time`` or
       ``pace`` is compiled. Variables that depend (in)directly on states are
       discounted.
    2. Each candidate is assign an initial score ``1 - 1 / (1 + d)``, where
       ``d`` is the distance (in the variable graph) to ``time`` or ``pace``.
       As a result, if we define ``t_beat = engine.time % 1000`` then this gets
       a lower score than ``I = if(t_beat > 100...)``.
    3. Candidates that have a unit compatible with current get +1 point.
    4. Candidates that define a unit (not ``None``) incompatible with current
       are discarded.
    5. Candidates whose name (case-insensitively) matches a known common
       stimulus current variable name get +2 points.
    6. Remaining candidates are ranked, and the highest scoring candidate is
       returned. If there is more than one highest scoring candidate an
       arbitrary one is selected.

    If no candiates are found with this approach, a third strategy is tried:

    1. All constants are scanned for variables with a known common name and a
       current unit (or None). The first such variable is returned.

    If all strategies fail, ``None`` is returned.
    """
    # Return any variable labelled as `stimulus_current`
    i_stim = model.label('stimulus_current')
    if i_stim is not None:
        return i_stim

    # Return any variable annotated as `oxmeta: membrane_stimulus_current`
    for v in model.variables(deep=True):
        if v.meta.get('oxmeta', None) == 'membrane_stimulus_current':
            return v

    # Units that current is often expressed in (give or take a multiplier)
    common_units = [
        myokit.units.A,
        myokit.units.A / myokit.units.F,
        myokit.units.A / myokit.units.m**2,
    ]

    # Non-dimensionalized (e.g. Mitchell-Schaeffer) use [mS/uF]
    # But this causes too many false positives
    # common_units.append(myokit.units.S / myokit.units.F)

    # Common names for stimulus current variables (lowercase)
    common_names = [
        'istim',
        'i_stim',
        'i_st',
        'ist',
    ]

    # Gather list of candidates and initial scores (based on distance to bound
    # variable)
    candidates = {}
    pace = model.binding('pace')
    if pace is not None:
        for var, distance in _distance_to_bound(pace).items():
            candidates[var] = 1 - 1 / (1 + distance)
        candidates.update()
    time = model.time()
    if time is not None:
        for var, distance in _distance_to_bound(time).items():
            candidates[var] = 1 - 1 / (1 + distance)

    # Add points for units (1 point if compatible with current), and filter out
    # any variables that explicitly specify a current incompatible unit
    incompatible = []
    for v in candidates:
        unit = v.unit()
        if unit is not None:
            if _compatible_units(unit, common_units):
                candidates[v] += 1
            else:
                incompatible.append(v)
    for v in incompatible:
        del candidates[v]

    # Add points for name (2 points if name matches common name)
    for v in candidates:
        name = v.name().lower()
        if name in common_names:
            candidates[v] += 2

    # Return candidates found with strategy 1
    if candidates:
        # Order from best to worst
        ranking = sorted(candidates.keys(), key=lambda v: -candidates[v])

        # Don't bother with tie-breaking, as long as something is returned!
        return ranking[0]

    # Back-up strategy: scan for constants with right name/unit
    for var in model.variables(const=True):
        name = var.name().lower()
        if name in common_names:
            unit = var.unit()
            if unit is None or _compatible_units(unit, common_units):
                return var
    return None


def stimulus_current_info(model):
    """
    Guesses the stimulus current variable and related variables representing
    the stimulus amplitude, duration, period, and offset.

    Returns a dict with entries ``current``, ``amplitude``, ``duration``,
    ``period``, ``offset``, and ``amplitude_expression``, any of which may be
    ``None``.

    The field ``amplitude_expression`` will be used if no variable representing
    the stimulus amplitude can be detected, but a literal expression for it can
    be found, e.g. from ``i_stim = -5 [pA]``.

    The method starts by guessing the stimulus current, and looking for
    annotations:

    1. The stimulus current is guessed using :meth:`stimulus_current()`. If no
       stimulus current is found the method terminates and returns a dict of
       ``None`` objects.
    2. Any variables annotated with oxmeta stimulus terms are found.

    If all fields (except ``amplitude_expression``) are filled now, the method
    returns. If not, the following strategy is used:

    1. The (direct and indirect) dependencies of the stimulus current are
       gathered in a list of candidate variables. Whenever a variable is used
       (e.g. as stimulus duration) it is removed from the list of candidates.
    2. The list is scanned for variables that don't declare a unit, or that
       declare a unit compatible with time. The first such variable containing
       the string ``duration`` (if any) is set as the duration variable, and
       the first containing either ``offset`` or ``start`` is used for the
       stimulus offset.
    3. The first candidate that has ``period`` in its name and has no units,
       time units, or is dimensionless, is used for the stimulus period.

    The stimulus amplitude is determined as follows:

    1. All remaining candidates are assigned a score ``1 / (1 + d)`` where
       ``d`` is the distance to the stimulus current variable (for a score in
       the range (0, 0.5]).
    2. Any candidates with unit other than None or the unit of the stimulus
       current are rejected.
    3. Candidates in the same unit as the stimulus current are given +1 points.
    4. Candidates with a name containing ``amplitude`` or ``current`` are given
       +1 points.
    5. If there are one or more candidates, the best is set as stimulus
       amplitude and the method returns.

    If the stimulus amplitude variable has not been determined, the following
    strategy is used to find a stimulus current _expression_:

    1. If the expression for the stimulus current is a constant, this is set as
       the amplitude expression.
    2. Otherwise, if the expression is a conditional (an if or a piecewise),
       then the first piece that isn't a literal which evaluates to zero is
       returned.

    At this point the method returns, with any undetermined fields left set to
    ``None``.
    """
    info = {
        'current': None,
        'amplitude': None,
        'duration': None,
        'period': None,
        'offset': None,
        'amplitude_expression': None,
    }

    # Guess stimulus current, return immediately if not found
    istim = stimulus_current(model)
    if istim is None:
        return info
    info['current'] = istim
    n_todo = 4

    # Check for oxmeta annotations
    used = set()
    oxmeta = [
        ('amplitude', 'membrane_stimulus_current_amplitude'),
        ('duration', 'membrane_stimulus_current_duration'),
        ('period', 'membrane_stimulus_current_period'),
        ('offset', 'membrane_stimulus_current_offset'),
    ]
    for v in model.variables(deep=True):
        ox = v.meta.get('oxmeta', None)
        if ox is not None:
            for k, annotation in oxmeta:
                if info[k] is None and ox == annotation:
                    info[k] = v
                    used.add(v)
                    n_todo -= 1

    # Stop if nothing left to do
    if n_todo == 0:
        return info

    # Time units
    time_units = [myokit.units.second]

    # Current unit: from stimulus current variable
    current_unit = istim.unit()

    # Gather dependencies of istim, and calculate distance-based score between
    # 0 and 0.5 (closer is better)
    candidates = {}
    for v, distance in _deep_deps(istim).items():
        if v not in used and v.binding() is None:
            candidates[v] = 1 / (1 + distance)

    # Guess time variables
    used = set()
    for v in candidates:

        # Find variables with potential time units
        unit = v.unit()
        good_unit = unit is None or _compatible_units(unit, time_units)

        # Duration & offset: must have time unit
        name = v.name().lower()
        if good_unit and info['duration'] is None:
            if 'duration' in name:
                info['duration'] = v
                used.add(v)
                n_todo -= 1
                continue
        if good_unit and info['offset'] is None:
            if 'offset' in name or 'start' in name:
                info['offset'] = v
                used.add(v)
                n_todo -= 1
                continue

        # Period is dimensionless is many models (not sure why)
        if good_unit or unit == myokit.units.dimensionless:
            if info['period'] is None and 'period' in name:
                info['period'] = v
                used.add(v)
                n_todo -= 1
                continue

    # Check if anything left to do
    if n_todo == 0:
        return info

    # Search remaining candidates for amplitude
    for v in used:
        del candidates[v]

    # Filter out incompatible units, award points for compatible ones
    rejected = set()
    for v in candidates:
        unit = v.unit()
        if unit is not None:
            if unit == current_unit:
                candidates[v] += 1
            else:
                rejected.add(v)
    for v in rejected:
        del candidates[v]

    # Award points for names
    for v in candidates:
        name = v.name().lower()
        if 'amplitude' in name or 'current' in name:
            candidates[v] += 1

    # Attempt to use highest ranked candidate
    if candidates:
        # Order from best to worst
        # Don't bother with tie-breaking, as long as something is returned!
        ranking = sorted(candidates.keys(), key=lambda v: -candidates[v])
        info['amplitude'] = ranking[0]
        return info

    # Back-up strategy: find an `amplitude_expression`

    # Option 1: Constant expression for i_stim
    e = istim.rhs()
    if e.is_constant():
        info['amplitude_expression'] = e

    # Option 2: Conditional expression (with condition depending on some
    # variable): find branch leading to a non-zero literal
    else:
        e = _find_conditional(e)
        if e is not None:
            if isinstance(e, myokit.If):
                e = e.piecewise()
            for piece in e.pieces():
                if not (piece.is_literal() and piece.eval() == 0):
                    info['amplitude_expression'] = piece
                    break

    return info




#
# Provides Matlab/Octave support
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from ._exporter import MatlabExporter
from ._ewriter import MatlabExpressionWriter


# Importers
# Exporters
_exporters = {
    'matlab': MatlabExporter,
}


def exporters():
    """
    Returns a dict of all exporters available in this module.
    """
    return dict(_exporters)


# Expression writers
_ewriters = {
    'matlab': MatlabExpressionWriter,
}


def ewriters():
    """
    Returns a dict of all expression writers available in this module.
    """
    return dict(_ewriters)


# Language keywords
keywords = [
    'i',
    'e',
    'pi'
]



#
# Provides support for the ChannelML format
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from ._importer import ChannelMLImporter, ChannelMLError  # noqa


# Importers
_importers = {
    'channelml': ChannelMLImporter,
}


def importers():
    """
    Returns a dict of all importers available in this module.
    """
    return dict(_importers)

# Exporters
# Expression writers



#
# Provides C++ support
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import myokit.formats.ansic

from ._ewriter import CppExpressionWriter


# Importers
# Exporters
# Expression writers
_ewriters = {
    'cpp': CppExpressionWriter,
}


def ewriters():
    """
    Returns a dict of all expression writers available in this module.
    """
    return dict(_ewriters)


# Language keywords
keywords = list(myokit.formats.ansic.keywords)
# TODO Append more keywords



# -*- coding: UTF-8 -*-
#
# OpenCL driven simulation, 1d or 2d
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import platform

from collections import OrderedDict

import numpy as np

import myokit


# Location of C and OpenCL sources
SOURCE_FILE = 'openclsim.c'
KERNEL_FILE = 'openclsim.cl'


class SimulationOpenCL(myokit.CModule):
    """
    Can run multi-cellular simulations based on a :class:`model <Model>` using
    OpenCL for parallelization.

    Takes the following input arguments:

    ``model``
        The model to simulate with. This model will be cloned when the
        simulation is created so that no changes to the given model will be
        made.
    ``protocol``
        An optional pacing protocol, used to stimulate a number of cells either
        at the start of a fiber or at the bottom-left of the tissue.
    ``ncells``
        The number of cells. Use a scalar for 1d simulations or a tuple
        ``(nx, ny)`` for 2d simulations.
    ``diffusion``
        Can be set to False to disable diffusion currents. This can be useful
        in combination with :meth:`set_field` to explore the effects of varying
        one or more parameters in a single cell model.
    ``precision``
        Can be set to ``myokit.SINGLE_PRECISION`` (default) or
        ``myokit.DOUBLE_PRECISION`` if the used device supports it.
    ``native_maths``
        On some devices, selected functions (e.g. ``exp``) can be made to run
        faster (but possibly less accurately) by setting ``native_maths=True``.
    ``rl``
        Use Rush-Larsen updates instead of forward Euler for any Hodgkin-Huxley
        gating variables (default=``False``).

    The simulation provides the following inputs variables can bind to:

    ``time``
        The simulation time
    ``pace``
        The pacing level, this is set if a protocol was passed in. Will be set
        to 0 if no protocol is provided.
    ``diffusion_current`` (if enabled)
        The current flowing from the cell to its neighbors. This will be
        positive when the cell is acting as a source, negative when it is
        acting as a sink.

    The input ``time`` is set globally: Any variable bound to ``time`` will
    appear in the logs as single, global variable (for example ``engine.time``
    instead of ``1.2.engine.time``. Variables bound to ``pace`` or
    ``diffusion_current`` are logged per cell. (If diffusion currents are
    disabled, the input ``diffusion_current`` will not be used, and any
    variables bound to it will be logged or not according to their default
    value).

    To set the number of cells that will be paced, the methods
    :meth:`set_paced_cells()` and :meth:`set_paced_cell_list()` can be used.

    A single labeled variable is required for this simulation to work:

    ``membrane_potential``
        The variable representing the membrane potential.

    Simulations maintain an internal state consisting of

    - the current simulation time
    - the current state
    - the default state

    When a simulation is created, the simulation time is set to 0 and both the
    current and the default state are equal to the state of the given model,
    copied once for each cell.
    After each call to :meth:`run` the time variable and current state are
    updated, so that each successive call to run continues where the previous
    simulation left off. A :meth:`reset` method is provided that will set the
    time back to 0 and revert the current state to the default state. To change
    the time or state manually, use :meth:`set_time` and :meth:`set_state`.

    A pre-pacing method :meth:`pre` is provided that doesn't affect the
    simulation time but will update the current *and the default state*. This
    allows you to pre-pace, run a simulation, reset to the pre-paced state, run
    another simulation etc.

    To set up a 1d simulation, the argument ``ncells`` should be given as a
    tuple. In this case, any cell ``i`` will be assumed to be connected to
    cells ``i - 1`` and ``i + 1`` (except at the boundaries).
    Similarly, if ``ncells`` is a 2-dimensional tuple ``(nx, ny)`` a grid will
    be assumed so that each (non-boundary) cell is connected to four
    neighbors.
    Finally, arbitrary geometries can be used by passing a scalar to ``ncells``
    and specifying the connections with :meth:`set_connections`.

    The ``diffusion_current`` in any cell is calculated as::

        i = sum[g * (V - V_j)]

    Where the sum is taken over all connected cells ``j`` (see [1]).

    Models used with this simulation need to have independent components: it
    should be possible to evaluate the model's equations one component at a
    time. A model's suitability can be tested using
    :meth:`has_interdependent_components
    <myokit.Model.has_interdependent_components>`.

    Simulations are performed using a forward Euler (FE) method with a fixed
    time step (see :meth:`set_step_size()`). Using smaller step sizes is
    computationally demanding, but gives more accurate results. Using too large
    a step size can also cause a simulation to become unstable, but please note
    that stability does imply accuracy, and always double-check important
    results by re-running with a reduced step size.

    If the optional parameter ``rl`` is set to ``True``, state variables
    written in a Hodgkin-Huxley form will be updated using a Rush-Larsen (RL)
    instead of a forward Euler step (see [2]). This provides greater stability
    (so that the step size can be increased) but not necessarily greater
    accuracy (see [3]), so that care must be taken when using this method.

    [1] Myokit: A simple interface to cardiac cellular electrophysiology.
    Clerx, Collins, de Lange, Volders (2016) Progress in Biophysics and
    Molecular Biology.

    [2] A practical algorithm for solving dynamic membrane equations.
    Rush, Larsen (1978) IEEE Transactions on Biomedical Engineering

    [3] Cellular cardiac electrophysiology modelling with Chaste and CellML
    Cooper, Spiteri, Mirams (2015) Frontiers in Physiology

    """
    _index = 0  # Unique id for the generated module

    def __init__(
            self, model, protocol=None, ncells=256, diffusion=True,
            precision=myokit.SINGLE_PRECISION, native_maths=False, rl=False):
        super().__init__()

        # Require a valid model
        if not model.is_valid():
            model.validate()

        # Require independent components
        if model.has_interdependent_components():
            cycles = '\n'.join([
                '  ' + ' > '.join([x.name() for x in c])
                for c in model.component_cycles()])
            raise ValueError(
                'This simulation requires models without interdependent'
                ' components. Please restructure the model and re-run.'
                '\nCycles:\n' + cycles)

        # Set protocol
        self.set_protocol(protocol)

        # Check dimensionality, number of cells
        try:
            self._nx = int(ncells)
            self._ny = 1
            self._dims = (self._nx,)
        except TypeError:
            try:
                assert len(ncells) == 2
            except (TypeError, AssertionError):
                raise ValueError(
                    'The argument "ncells" must be either a scalar or a tuple'
                    ' (nx, ny).')
            self._nx = int(ncells[0])
            self._ny = int(ncells[1])
            self._dims = (self._nx, self._ny)
        if self._nx < 1 or self._ny < 1:
            raise ValueError(
                'The number of cells in any direction must be at least 1.')
        self._ntotal = self._nx * self._ny

        # Set diffusion mode
        self._diffusion_enabled = bool(diffusion)

        # Set precision
        if precision not in (myokit.SINGLE_PRECISION, myokit.DOUBLE_PRECISION):
            raise ValueError('Only single and double precision are supported.')
        self._precision = precision

        # Set native maths
        self._native_math = bool(native_maths)

        # Set rush-larsen mode
        self._rl = bool(rl)

        # Get membrane potential variable (from pre-cloned model!)
        vm = model.label('membrane_potential')
        if self._diffusion_enabled or self._rl:
            if vm is None:
                raise ValueError(
                    'This simulation requires the membrane potential'
                    ' variable to be labelled as "membrane_potential".')
            if not vm.is_state():
                raise ValueError(
                    'The variable labelled as membrane potential must'
                    ' be a state variable.')

        #if vm.is_referenced():
        #  raise ValueError('This simulation requires that no other variables'
        #      ' depend on the time-derivative of the membrane potential.')

        # Prepare for Rush-Larsen updates, and/or clone model
        self._rl_states = {}
        if self._rl:
            import myokit.lib.hh as hh

            # Convert alpha-beta formulations to inf-tau forms, cloning model
            self._model = hh.convert_hh_states_to_inf_tau_form(model, vm)
            self._vm = self._model.get(vm.qname())
            del model, vm

            # Get (inf, tau) tuple for every Rush-Larsen state
            for state in self._model.states():
                res = hh.get_inf_and_tau(state, self._vm)
                if res is not None:
                    self._rl_states[state] = res

        else:
            # Clone model, store
            self._model = model.clone()
            if vm is None:
                self._vm = None
            else:
                self._vm = self._model.get(vm.qname())
            del model, vm

        # Set default conductance values
        self._gx = self._gy = None
        if self._diffusion_enabled:
            self.set_conductance()

        # Set conductance fields
        self._gx_field = None
        self._gy_field = None

        # Set connections
        self._connections = None

        # Set default paced cells
        self._paced_cells = None
        if self._diffusion_enabled:
            self.set_paced_cells()

        # Scalar fields
        self._fields = OrderedDict()

        # Set default time step
        self.set_step_size()

        # Set initial time
        self._time = 0

        # Count number of states
        self._nstate = self._model.count_states()

        # Set state and default state
        self._state = self._model.initial_values(True) * self._ntotal
        self._default_state = list(self._state)

        # List of globally logged inputs
        self._global = ['time', 'pace']

        # Process bindings: remove unsupported bindings, get map of bound
        # variables to internal names.
        inputs = {'time': 'time', 'pace': 'pace'}
        if self._diffusion_enabled:
            inputs['diffusion_current'] = 'idiff'
        self._bound_variables = myokit._prepare_bindings(self._model, inputs)

        # Create unique names
        self._model.create_unique_names()

        # Create back-end
        SimulationOpenCL._index += 1
        mname = 'myokit_sim_opencl_' + str(SimulationOpenCL._index)
        mname += '_' + str(myokit.pid_hash())
        fname = os.path.join(myokit.DIR_CFUNC, SOURCE_FILE)
        args = {
            'module_name': mname,
            'model': self._model,
            'precision': self._precision,
            'dims': len(self._dims),
        }

        # Define libraries
        libs = []
        flags = []
        plat = platform.system()
        if plat != 'Darwin':    # pragma: no macos cover
            libs.append('OpenCL')
        else:                   # pragma: no cover
            flags.append('-framework')
            flags.append('OpenCL')
        if plat != 'Windows':   # pragma: no windows cover
            libs.append('m')

        # Create extension
        libd = list(myokit.OPENCL_LIB)
        incd = list(myokit.OPENCL_INC)
        incd.append(myokit.DIR_CFUNC)
        self._sim = self._compile(
            mname, fname, args, libs, libd, incd, larg=flags,
            continue_in_debug_mode=True)

    def calculate_conductance(self, r, sx, chi, dx):
        """
        This method is deprecated, please use :meth:`monodomain_conductance`
        instead, but note the difference in the arguments::

            calculate_conductance(r, sx, chi, dx)

        is equivalent to::

            monodomain_conductance(1 / chi, r, sx, dx, 1)

        """
        # Deprecated since 2021-06-25
        import warnings
        warnings.warn('The method SimulationOpenCL.calculate_conductance() is'
                      ' deprecated. Please use monodomain_conductance()'
                      ' instead.')
        return self.monodomain_conductance(1 / chi, r, sx, dx, 1)

    def conductance(self):
        """
        Returns the cell-to-cell conductance(s) used in this simulation.

        The returned value will be a single float for 1d simulations, a tuple
        ``(gx, gy)`` for 2d simulations, and ``None`` if conductances were set
        with :meth:`set_conductance_field` or :meth:`set_connections()`.

        If diffusion is disabled, a call to this method will raise a
        ``RuntimeError``.
        """
        if not self._diffusion_enabled:
            raise RuntimeError(
                'This method is unavailable when diffusion is disabled.')
        if (self._gx_field is not None) or (self._connections is not None):
            return None
        if len(self._dims) == 1:
            return self._gx
        return (self._gx, self._gy)

    def default_state(self, x=None, y=None):
        """
        Returns the current default simulation state as a list of
        ``len(state) * n_total_cells`` floating point values, where
        ``n_total_cells`` is the total number of cells.

        If the optional arguments ``x`` and ``y`` specify a valid cell index a
        single cell's state is returned. For example ``state(4)`` can be
        used with a 1d simulation, while ``state(4, 2)`` is a valid index in
        the 2d case.

        For 2d simulations, the list is indexed so that x changes first (the
        first ``nx`` entries have ``y = 0``, the second ``nx`` entries have
        ``y = 1`` and so on).
        """
        if x is None:
            return list(self._default_state)
        else:
            x = int(x or 0)
            if x < 0 or x >= self._nx:
                raise IndexError('Given x-index out of range.')
            y = int(y or 0)
            if len(self._dims) == 2:
                if y < 0 or y >= self._ny:
                    raise IndexError('Given y-index out of range.')
                x += y * self._nx
            elif y != 0:
                raise ValueError(
                    'Y-coordinate specified for 1-dimensional simulation.')

            return self._default_state[x * self._nstate:(x + 1) * self._nstate]

    def find_nan(self, log, watch_var=None, safe_range=None, return_log=False):
        """
        Searches for the origin of a bad value (``NaN`` or ``inf``) in a data
        log generated by this simulation.

        Arguments:

        ``log``
            A :class:`myokit.DataLog` from this simulation. The log must
            contain the state of each cell and all bound variables. The bad
            value can occur at any point in time except the first.
        ``watch_var``
            To aid in diagnosis, a variable can be selected as ``watch_var``
            and a ``safe_range`` can be specified. With this option, the
            function will find and report either the first bad value or the
            first time the watched variable left the safe range, whatever came
            first. The watched variable must be a state variable.
        ``safe_range``
            The safe range to check the ``watch_var`` against. The safe range
            should be specified as ``(lower, upper)`` where both bounds are
            assumed to be in the safe range.
        ``return_log``
            If set to ``True``, a log containing the final points before the
            error occurred will be returned, containing all variables for all
            cells.

        Returns a tuple containing the following six values (or seven values if
        ``return_log=True``):

        ``time``
            The time the first bad value was found.
        ``icell``
            The index of the cell in which the bad value was found, as a tuple
            e.g. ``(3, )`` in a 1d simulation or ``(15, 12)`` in a 2d
            simulation.
        ``variable``
            The name of the variable that was detected to be ``NaN`` or
            ``inf``.
        ``value``
            The bad value detected)
        ``states``
            The state at time ``time`` and, if possible, up to 3 earlier
            states. Here, ``states[0]`` points to the current state,
            ``state[1]`` is the previous state and so on. Each state is
            represented as a list of values.
        ``bounds``
            The bound variables corresponding to the returned ``states``. For
            ``diffusion_current``, this will contain the specific cell's
            current, for ``time`` and ``pace`` the global values are shown.
            Each entry in ``bounds`` is a dictionary from variable names to
            values.
        ``log``
            A :class:`myokit.DataLog` with all variables for the same points as
            ``states`` and ``bounds``. This will only be included if
            ``return_log=True``.

        """
        import numpy as np

        # Test if log contains all states and bound variables
        t = []
        for label in self._global:
            var = self._model.binding(label)
            if var is not None:
                t.append(var.qname())
        t = myokit.prepare_log(
            myokit.LOG_STATE + myokit.LOG_BOUND,
            self._model,
            dims=self._dims,
            global_vars=t)
        for key in t:
            if key not in log:
                raise myokit.FindNanError(
                    'Method requires a simulation log containing all states'
                    ' and bound variables. Missing variable <' + key + '>.')
        del t

        # Error criterium
        if watch_var is None:

            # NaN/inf detection
            def bisect(ar, lo, hi):
                if not np.isfinite(ar[lo]):
                    return lo
                md = lo + int(np.ceil(0.5 * (hi - lo)))
                if md == hi:
                    return hi
                if not np.isfinite(ar[md]):
                    return bisect(ar, lo, md)
                else:
                    return bisect(ar, md, hi)

            def find_error_position(log):
                # Search for first occurrence of propagating NaN in the log
                ifirst = None   # Index in time
                kfirst = None   # Variable + cell index
                for key, ar in log.items():
                    if ifirst is None:
                        if not np.isfinite(ar[-1]):
                            # NaN found in the log
                            kfirst = key
                            ifirst = bisect(ar, 0, len(ar) - 1)
                            if ifirst == 0:
                                break
                    elif not np.isfinite(ar[ifirst - 1]):   # pragma: no cover
                        # Earlier NaN found than before
                        kfirst = key
                        ifirst = bisect(ar, 0, ifirst)
                        if ifirst == 0:
                            break
                return ifirst, kfirst

        else:

            # Variable out of bounds detection
            try:
                watch_var = self._model.get(watch_var)
            except KeyError:
                raise myokit.FindNanError(
                    'Variable <' + str(watch_var) + '> not found.')
            if not watch_var.is_state():
                raise myokit.FindNanError(
                    'The watched variable must be a state.')
            try:
                lo, hi = safe_range
            except Exception:
                raise myokit.FindNanError(
                    'A safe range must be specified for the watched variable'
                    ' as a tuple (lower, upper).')
            if lo >= hi:
                raise myokit.FindNanError(
                    'The safe range must have a lower bound that is lower than'
                    ' the upper bound.')

            def find_error_position(_log):
                # Find first occurence of out-of-bounds error
                ifirst = None
                kfirst = None
                post = '.' + watch_var.qname()
                lower, upper = safe_range
                for dims in myokit._dimco(*self._dims):
                    key = '.'.join([str(x) for x in dims]) + post
                    ar = np.array(_log[key], copy=False)
                    i = np.where(
                        (ar < lower)
                        | (ar > upper)
                        | np.isnan(ar)
                        | np.isinf(ar))[0]
                    if len(i) > 0:
                        i = i[0]
                        if ifirst is None:
                            kfirst = key
                            ifirst = i
                        elif i < ifirst:
                            kfirst = key
                            ifirst = i
                        if i == 0:  # pragma: no cover
                            break
                return ifirst, kfirst

        # Get the name of a time variable
        time_var = self._model.time().qname()

        # Deep searching function
        def relog(_log, _dt):
            # Get first occurence of error
            ifirst, kfirst = find_error_position(_log)
            if kfirst is None:
                raise myokit.FindNanError('Error condition not found in log.')
            if ifirst == 0:
                raise myokit.FindNanError(
                    'Unable to work with simulation logs where the error'
                    ' condition is met in the very first data point.')

            # Position to start deep search at
            istart = ifirst - 1

            # Get last logged state before error
            state = []
            for dims in myokit._dimco(*self._dims):
                pre = '.'.join([str(x) for x in dims]) + '.'
                for s in self._model.states():
                    state.append(_log[pre + s.qname()][istart])

            # Get last time before error
            time = _log[time_var][istart]

            # Save current state & time
            old_state = self._state
            old_time = self._time
            self._state = state
            self._time = time

            # Run until next time point, log every step
            duration = _log[time_var][ifirst] - time
            _log = self.run(
                duration, log=myokit.LOG_BOUND + myokit.LOG_STATE,
                log_interval=_dt, report_nan=False)

            # Reset simulation to original state
            self._state = old_state
            self._time = old_time

            # Return new log
            return _log

        # Get time step
        try:
            dt = log[time_var][1] - log[time_var][0]
        except IndexError:  # pragma: no cover
            # Unable to guess dt!
            # So... Nan occurs before the first log interval is reached
            # That probably means dt was relatively large, so guess it was
            # large! Assuming milliseconds, start off with dt=5ms
            dt = 5
            # Note: This seems impossible to reach at the moment, but would be
            # possible if a log_times argument was supported.

        # Search with successively fine log interval
        while dt > 0:
            dt *= 0.1
            if dt < 0.5:
                dt = 0
            log = relog(log, dt)

        # Search for first occurrence of error in the detailed log
        ifirst, kfirst = find_error_position(log)
        if ifirst is None:
            ifirst = 0
            kfirst = next(iter(log.items()))[0]

        # Get indices of cell in state vector
        ndims = len(self._dims)
        icell = [int(x) for x in kfirst.split('.')[0:ndims]]

        # Get state & bound before, during and after error
        def state(index, icell):
            s = []
            b = {}
            for var in self._model.states():
                s.append(log[var.qname(), icell][index])
            for var in self._model.variables(bound=True):
                if var.binding() in self._global:
                    b[var.qname()] = log[var.qname()][index]
                else:
                    b[var.qname()] = log[var.qname(), icell][index]
            return s, b

        # Get error cell's states before, during and after
        times = []
        states = []
        bounds = []
        max_states = 3
        for k in range(ifirst, ifirst - max_states - 1, -1):
            if k < 0:
                break
            times.append(log[time_var][k])
            s, b = state(k, icell)
            states.append(s)
            bounds.append(b)

        # Get variable causing error
        var = self._model.get('.'.join(kfirst.split('.')[ndims:]))

        # Get value causing error
        if var.is_state():
            value = states[1 if ifirst > 0 else 0][var.index()]
        else:  # pragma: no cover
            value = bounds[1 if ifirst > 0 else 0][var.qname()]
        var = var.qname()

        # Get time error occurred
        time = log[time_var][ifirst]

        # Get all variables at shown states
        if return_log:
            # Get earliest state in states/bounds
            state = []
            istart = max(0, ifirst - max_states)
            for dims in myokit._dimco(*self._dims):
                pre = '.'.join([str(x) for x in dims]) + '.'
                for s in self._model.states():
                    state.append(log[pre + s.qname()][istart])

            # Save current state & time, and rewind
            old_state = self._state
            old_time = self._time
            self._state = state
            self._time = times[-1]

            # Run for all states
            duration = len(times) * self._step_size
            log = self.run(
                duration,
                log=myokit.LOG_BOUND + myokit.LOG_STATE + myokit.LOG_INTER,
                log_interval=0,
                report_nan=False)

            # Reset simulation to original state
            self._state = old_state
            self._time = old_time

            # Return
            return time, icell, var, value, states, bounds, log

        return time, icell, var, value, states, bounds

    def is2d(self):
        """Deprecated alias of :meth:`is_2d()`."""
        # Deprecated since 2020-09-10
        import warnings
        warnings.warn('The method SimulationOpenCL.is2d() is deprecated.'
                      ' Please use is_2d() instead.')
        return self.is_2d()

    def is_2d(self):
        """Returns ``True`` if and only if this is a 2d simulation. """
        return len(self._dims) == 2

    def is_paced(self, x, y=None):
        """
        Returns ``True`` if and only if the cell at index ``x`` (or index
        ``(x, y)`` in 2d simulations) will be paced during simulations.

        If diffusion is disabled, a call to this method will raise a
        ``RuntimeError``.
        """
        if not self._diffusion_enabled:
            raise RuntimeError(
                'This method is unavailable when diffusion is disabled.')

        # Check input
        x = int(x)
        if x < 0 or x >= self._dims[0]:
            raise IndexError('X-coordinate out of range: ' + str(x) + '.')
        if len(self._dims) == 2:
            if y is None:
                raise ValueError(
                    'No y-coordinate specified in 2-dimensional simulation.')
            y = int(y)
            if y < 0 or y >= self._dims[1]:
                raise IndexError('Y-coordinate out of range: ' + str(y) + '.')
        elif y is None:
            y = 0
        elif y != 0:
            raise ValueError(
                'Y-coordinate specified for 1-dimensional simulation.')

        # Pacing rectangle
        if type(self._paced_cells) == tuple:
            nx, ny, xmin, ymin = self._paced_cells
            return (x >= xmin and x < xmin + nx and
                    y >= ymin and y < ymin + ny)

        # Explicit cell selection
        cid = x + y * self._dims[0]
        return cid in self._paced_cells

    def monodomain_conductance(self, chi, k, D, dx, A=1):
        """
        Calculates conductance values ``g`` based on monodomain parameters,
        assuming that membrane size, capacitance, and cell-to-cell conductance
        do not vary spatially.

        In this simulation, the membrane potential ``V`` of a cell is assumed
        to be given by::

            dV/dt = -1/Cm (I_ion + I_stim + I_diff)

        where Cm is membrane capacitance and the diffusion current is defined
        by::

            I_diff[i] = sum[g[i,j] * (V[i] - V[j])]

        in which the sum is over all neighbors ``j`` of cell ``i``.

        Alternatively, with capacitance and currents normalized to membrane
        area, we can write::

            dV/dt = -1/cm (i_ion + i_stim + i_diff)

        In the monodomain model (see e.g. [4]), this diffusion current per unit
        area is defined as::

            i_diff = -(1 / chi) (k / (k + 1)) ∇ * (D∇V)

        (see the argument list below for the meaning of the variables).

        This can be equated to Myokit's diffusion current, but only if we
        assume *zero-flux boundary conditions*, a *regularly spaced grid*,
        and *no spatial heterogeneity in D* (or g).

        With these assumptions, we can use finite differences to find::

            g_bar = (1 / chi) * (k / (k + 1)) * D * (1 / dx^2)

        where ``g_bar`` is the cell-to-cell conductance, but normalized with
        respect to unit membrane area.
        For models with currents normalized to area this is unproblematic, but
        to convert to models with unnormalized currents this means we have
        added the further assumption that *each node contains some fixed
        amount of membrane*, determined by an area A::

            g = (1 / chi) * (k / (k + 1)) * D * (1 / dx^2) * A

        This equation can also be applied in two dimensions, but only if
        *we assume that the conductivity matrix is diagonal*, in which case::

            gx = (1 / chi) * (k / (k + 1)) * Dx * (1 / dx^2) * A
            gy = (1 / chi) * (k / (k + 1)) * Dy * (1 / dy^2) * A

        This method uses the above equation to calculate and return a
        conductance value from the parameters used in monodomain simulations.

        Arguments:

        ``chi``
            The surface area of the membrane per unit volume (in units
            area per volume).
        ``k``
            The intra- to extracellular conductivity ratio.
        ``D``
            The intracellular conductivity (in siemens per unit length).
        ``dx``
            The distance between grid points.
        ``A``
            The area of the cell membrane.

        [4] Computer Model of Excitation and Recovery in the Anisotropic
        Myocardium I. Rectangular and Cubic Arrays of Excitable Elements.
        Leon & Horacek (1991) Journal of electrocardiology
        """
        return D * A * k / ((k + 1) * chi * dx * dx)

    def neighbors(self, x, y=None):
        """
        Returns a list of indices specifying the neighbors of the cell at
        index ``x`` (or index ``(x, y)`` for 2d simulations).

        Indices are given either as integers (1d or arbitrary geometry) or as
        tuples (2d).

        If diffusion is disabled, a call to this method will raise a
        ``RuntimeError``.
        """
        if not self._diffusion_enabled:
            raise RuntimeError(
                'This method is unavailable when diffusion is disabled.')

        # Check input
        x = int(x or 0)
        if x < 0 or x >= self._dims[0]:
            raise IndexError('X-coordinate out of range: ' + str(x) + '.')
        if len(self._dims) == 2:
            if y is None:
                raise ValueError(
                    'No y-coordinate specified in 2-dimensional simulation.')
            y = int(y)
            if y < 0 or y >= self._dims[1]:
                raise IndexError('Y-coordinate out of range: ' + str(y) + '.')
        elif not (y is None or y == 0):
            raise ValueError(
                'Y-coordinate specified for 1-dimensional simulation.')

        # User-specified connections (always 1d)
        if self._connections is not None:
            neighbors = []
            for i, j, c in self._connections:
                if i == x:
                    neighbors.append(j)
                elif j == x:
                    neighbors.append(i)
            return neighbors

        # Left and right neighbors
        neighbors = []
        if x > 0:
            neighbors.append(x - 1)
        if x + 1 < self._dims[0]:
            neighbors.append(x + 1)

        # Top and bottom neighbors
        if len(self._dims) == 2:
            neighbors = [(i, y) for i in neighbors]
            if y > 0:
                neighbors.append((x, y - 1))
            if y + 1 < self._dims[1]:
                neighbors.append((x, y + 1))
        return neighbors

    def neighbours(self, x, y=None):
        """ Alias of :meth:`neighbors`. """
        # Deprecated since 2023-06-07
        import warnings
        warnings.warn('The method SimulationOpenCL.neighbours() is deprecated.'
                      ' Please use neighbors() instead.')

        return self.neighbors(x, y)

    def pre(self, duration, report_nan=True, progress=None,
            msg='Pre-pacing SimulationOpenCL'):
        """
        This method can be used to perform an unlogged simulation, typically to
        pre-pace to a (semi-)stable orbit.

        After running this method

        - The simulation time is **not** affected
        - The current state and the default state are updated to the final
          state reached in the simulation.

        Calls to :meth:`reset` after using :meth:`pre` will revert the
        simulation to this new default state.

        If numerical errors during the simulation lead to NaNs appearing in the
        result, the ``find_nan`` method will be used to pinpoint their
        location. Next, a call to the model's rhs will be evaluated in python
        using checks for numerical errors to pinpoint the offending equation.
        The results of these operations will be written to ``stdout``. To
        disable this feature, set ``report_nan=False``.

        To obtain feedback on the simulation progress, an object implementing
        the :class:`myokit.ProgressReporter` interface can be passed in.
        passed in as ``progress``. An optional description of the current
        simulation to use in the ProgressReporter can be passed in as `msg`.
        """
        self._run(duration, myokit.LOG_NONE, 1, report_nan, progress, msg)
        self._default_state = list(self._state)

    def remove_field(self, var):
        """
        Removes any field set for the given variable.
        """
        if isinstance(var, myokit.Variable):
            var = var.qname()
        var = self._model.get(var)
        del self._fields[var]

    def reset(self):
        """
        Resets the simulations:

        - The time variable is set to 0
        - The current state is set to the default state (either the model's
          initial state or the last state reached using :meth:`pre`)

        """
        self._time = 0
        self._state = list(self._default_state)

    def run(self, duration, log=None, log_interval=1.0, report_nan=True,
            progress=None, msg='Running SimulationOpenCL'):
        """
        Runs a simulation and returns the logged results. Running a simulation
        has the following effects:

        - The internal state is updated to the last state in the simulation.
        - The simulation's time variable is updated to reflect the time
          elapsed during the simulation.

        The number of time units to simulate can be set with ``duration``.

        The variables to log can be indicated using the ``log`` argument. There
        are several options for its value:

        - ``None`` (default), to log all states
        - An integer flag or a combination of flags. Options:
          ``myokit.LOG_NONE``, ``myokit.LOG_STATE``, ``myokit.LOG_BOUND``,
          ``myokit.LOG_INTER`` or ``myokit.LOG_ALL``.
        - A list of qnames or variable objects
        - A :class:`myokit.DataLog` object or another dictionary of
           ``qname : list`` mappings.

        For more details on the ``log`` argument, see the function
        :meth:`myokit.prepare_log`.

        Variables that vary from cell to cell will be logged with a prefix
        indicating the cell index. For example, when using::

            s = SimulationOpenCL(m, p, ncells=256)
            d = s.run(1000, log=['engine.time', 'membrane.V']

        where ``engine.time`` is bound to ``time`` and ``membrane.V`` is the
        membrane potential variable, the resulting log will contain the
        following variables::

            {
                'engine.time'  : [...],
                '0.membrane.V' : [...],
                '1.membrane.V' : [...],
                '2.membrane.V' : [...],
            }

        Alternatively, you can specify variables exactly::

            d = s.run(1000, log=['engine.time', '0.membrane.V']

        For 2d simulations, the naming scheme ``x.y.name`` is used, for
        example ``0.0.membrane.V``.

        A log entry will be made every time *at least* ``log_interval`` time
        units have passed. No guarantee is given about the exact time log
        entries will be made, but the value of any logged time variable is
        guaranteed to be accurate.

        If numerical errors during the simulation lead to NaNs appearing in the
        result, the ``find_nan`` method will be used to pinpoint their
        location. Next, a call to the model's rhs will be evaluated in python
        using checks for numerical errors to pinpoint the offending equation.
        The results of these operations will be written to ``stdout``. To
        disable this feature, set ``report_nan=False``.

        To obtain feedback on the simulation progress, an object implementing
        the :class:`myokit.ProgressReporter` interface can be passed in.
        passed in as ``progress``. An optional description of the current
        simulation to use in the ProgressReporter can be passed in as `msg`.
         """
        r = self._run(duration, log, log_interval, report_nan, progress, msg)
        self._time += duration
        return r

    def _run(self, duration, log, log_interval, report_nan, progress, msg):
        # Simulation times
        if duration < 0:
            raise ValueError('Simulation time can\'t be negative.')
        tmin = self._time
        tmax = tmin + duration

        # Gather global variables in model
        g = []
        for label in self._global:
            v = self._model.binding(label)
            if v is not None:
                g.append(v.qname())

        # Parse log argument
        log = myokit.prepare_log(
            log,
            self._model,
            dims=self._dims,
            global_vars=g,
            if_empty=myokit.LOG_STATE + myokit.LOG_BOUND,
            allowed_classes=myokit.LOG_STATE + myokit.LOG_INTER
            + myokit.LOG_BOUND,
            precision=self._precision)

        # Create list of intermediary variables that need to be logged
        inter_log = []
        vars_checked = set()
        for var in log.keys():
            var = myokit.split_key(var)[1]
            if var in vars_checked:
                continue
            vars_checked.add(var)
            var = self._model.get(var)
            if var.is_intermediary() and not var.is_bound():
                inter_log.append(var)

        # Get preferred platform/device combo from configuration file
        platform, device = myokit.OpenCL.load_selection_bytes()

        # Compile template into string with kernel code
        kernel_file = os.path.join(myokit.DIR_CFUNC, KERNEL_FILE)
        args = {
            'model': self._model,
            'precision': self._precision,
            'native_math': self._native_math,
            'bound_variables': self._bound_variables,
            'inter_log': inter_log,
            'diffusion': self._diffusion_enabled,
            'fields': self._fields.keys(),
            'paced_cells': self._paced_cells,
            'rl_states': self._rl_states,
            'connections': self._connections is not None,
            'heterogeneous': self._gx_field is not None,
            'fiber_tissue': False,
        }
        kernel = self._export(kernel_file, args)

        # Logging period (0 = disabled)
        log_interval = 1e-9 if log_interval is None else float(log_interval)
        if log_interval <= 0:
            log_interval = 1e-9

        # Create field values vector
        n = len(self._fields) * self._nx * self._ny
        if n:
            field_data = self._fields.values()
            field_data = [np.array(x, copy=False) for x in field_data]
            field_data = np.vstack(field_data)
            field_data = list(field_data.reshape(n, order='F'))
        else:
            field_data = []

        # Get progress indication function (if any)
        if progress is None:
            progress = myokit._Simulation_progress
        if progress:
            if not isinstance(progress, myokit.ProgressReporter):
                raise ValueError(
                    'The argument "progress" must be either a subclass of'
                    ' myokit.ProgressReporter or None.')

        # Run simulation
        arithmetic_error = False
        if duration > 0:
            # Initialize
            state_in = self._state
            state_out = list(state_in)
            self._sim.sim_init(
                platform,
                device,
                kernel,
                self._nx,
                self._ny,
                self._diffusion_enabled,
                self._gx or 0,
                self._gy or 0,
                self._gx_field,
                self._gy_field,
                self._connections,
                tmin,
                tmax,
                self._step_size,
                state_in,
                state_out,
                self._protocol,
                log,
                log_interval,
                [x.qname().encode('ascii') for x in inter_log],
                field_data,
            )
            t = tmin
            try:
                if progress:
                    # Loop with feedback
                    with progress.job(msg):
                        r = 1.0 / duration if duration != 0 else 1
                        while t < tmax:
                            t = self._sim.sim_step()
                            if not progress.update(min((t - tmin) * r, 1)):
                                raise myokit.SimulationCancelledError()
                else:
                    # Loop without feedback
                    while t < tmax:
                        t = self._sim.sim_step()
            except ArithmeticError:
                arithmetic_error = True
            finally:
                # Clean even after KeyboardInterrupt or other Exception
                self._sim.sim_clean()
            # Update state
            self._state = state_out

        # Check for NaN
        if report_nan and (arithmetic_error or log.has_nan()):
            txt = ['Numerical error found in simulation logs.']
            try:
                # NaN encountered, show how it happened
                time, icell, var, value, states, bounds, d = self.find_nan(
                    log, return_log=True)
                icell_str = '(' + ','.join([str(x) for x in icell]) + ')'
                txt.append(
                    'Encountered numerical error at t='
                    + myokit.float.str(time, precision=self._precision)
                    + ' in cell ' + icell_str + ' when ' + var + '='
                    + myokit.float.str(value, precision=self._precision) + '.')

                # Check if this cell was paced.
                is_paced = self.is_paced(*icell)
                is_paced_str = (', cell ' + icell_str + ' '
                                + ('IS' if is_paced else 'is NOT') + ' paced')

                # Get the names of the variables used in bindings, to use when
                # calculating the derivatives
                vtime = self._model.time().qname()
                vpace = self._model.binding('pace')
                vpace = None if vpace is None else vpace.qname()
                vdiff = None
                if self._diffusion_enabled:
                    vdiff = self._model.binding('diffusion_current')
                    vdiff = None if vdiff is None else vdiff.qname()

                # List all neighbors for this cell
                neighbors = self.neighbors(*icell)
                if len(self._dims) == 2:
                    neighbors_str = ', '.join(
                        ('(' + ','.join([str(i) for i in j]) + ')')
                        for j in neighbors)
                else:
                    neighbors_str = ', '.join(str(j) for j in neighbors)

                # Show final state
                txt.append('State during:')
                txt.append(self._model.format_state(
                    states[0], precision=self._precision))

                # Show bound variables
                txt.append('Simulation variables during:')
                txt.append('  Time: ' + myokit.float.str(
                    bounds[0][vtime], precision=self._precision))
                if vpace is not None:
                    txt.append(
                        '  Pacing variable: ' + myokit.float.str(
                            bounds[0][vpace], precision=self._precision)
                        + is_paced_str)
                if vdiff is not None:
                    txt.append(
                        '  Diffusion current: ' + myokit.float.str(
                            bounds[0][vdiff], precision=self._precision))
                if neighbors:
                    txt.append('  Connected cells: ' + neighbors_str)

                # Show previous state (and derivatives)
                n_states = len(states)
                txt.append('Obtained ' + str(n_states) + ' previous state(s).')
                if n_states > 1:
                    # Get state and input at previous state
                    state = states[1]
                    bound = {
                        'time': bounds[1][vtime],
                        'pace': 0,
                        'diffusion_current': 0,
                    }
                    if is_paced and vpace is not None:
                        bound['pace'] = bounds[1][vpace]
                    if vdiff is not None:
                        bound['diffusion_current'] = bounds[1][vdiff]

                    # Evaluate state derivatives
                    eval_error = None
                    try:
                        derivs = self._model.evaluate_derivatives(
                            state, bound, self._precision, ignore_errors=False)
                    except myokit.NumericalError as ee:
                        derivs = self._model.evaluate_derivatives(
                            state, bound, self._precision, ignore_errors=True)
                        eval_error = str(ee)

                    # Show state and derivatives
                    txt.append('State before:')
                    txt.append(self._model.format_state_derivatives(
                        state, derivs, self._precision))

                    # Show bound variables
                    txt.append('Simulation variables before:')
                    txt.append('  Time: ' + myokit.float.str(
                        bounds[1][vtime], precision=self._precision))
                    if vpace is not None:
                        txt.append(
                            '  Pacing variable: ' + myokit.float.str(
                                bounds[1][vpace], precision=self._precision)
                            + is_paced_str)
                    if vdiff is not None:
                        txt.append(
                            '  Diffusion current: ' + myokit.float.str(
                                bounds[1][vdiff], precision=self._precision))
                    if neighbors:
                        txt.append('  Connected cells: ' + neighbors_str)

                    # Show all variables with non-finite values
                    txt.append(
                        'Logged all variables for points: ' + ', '.join(
                            myokit.float.str(t, precision=self._precision)
                            for t in d.time()))
                    txt.append(
                        'Non-finite valued variables at t=' + myokit.float.str(
                            d.time()[-2], precision=self._precision))
                    for key, values in d.items():
                        value = values[-2]
                        if not np.isfinite(value):
                            txt.append(
                                '  ' + str(key) + ' = ' + myokit.float.str(
                                    value, precision=self._precision))

                    # Show any error in evaluating derivatives
                    if eval_error is not None:
                        txt.append('Error when evaluating derivatives:')
                        txt.append(eval_error)

            except myokit.FindNanError as e:
                txt.append(
                    'Unable to pinpoint source of NaN, an error occurred:')
                txt.append(str(e))
            raise myokit.SimulationError('\n'.join(txt))

        # Return log
        return log

    def set_conductance(self, gx=10, gy=5):
        """
        Sets the cell-to-cell conductance used in this simulation.

        For 1d simulations, only ``gx`` will be used and the argument ``gy``
        can be omitted. For 2d simulations both arguments should be set.

        The diffusion current is calculated as::

            i = gx * ((V - V_xnext) - (V_xlast - V))
              + gy * ((V - V_ynext) - (V_ylast - V))

        Where the second term ``gy * ...`` is only used for 2d simulations. At
        the boundaries, where either ``V_ilast`` or ``V_inext`` is unavailable,
        the value of ``V`` is substituted, causing the term to go to zero.

        For a model with currents in ``[uA/uF]`` and voltage in ``[mV]``,
        `gx`` and ``gy`` have the unit ``[mS/uF]``.

        Calling ``set_conductance`` will delete any conductances previously set
        with :meth:`set_conductance_field` or :meth:`set_connections`.

        If diffusion is disabled, a call to this method will raise a
        ``RuntimeError``.
        """
        if not self._diffusion_enabled:
            raise RuntimeError(
                'This method is unavailable when diffusion is disabled.')

        gx, gy = float(gx), float(gy)
        if gx < 0:
            raise ValueError('Invalid conductance gx=' + str(gx))
        if gy < 0:
            raise ValueError('Invalid conductance gy=' + str(gy))
        self._gx, self._gy = gx, gy
        self._gx_field = self._gy_field = None
        self._connections = None

    def set_conductance_field(self, gx, gy=None):
        """
        Sets the cell-to-cell conductances used in this simulation, using lists
        of conductances.

        For 1d simulations, the argument ``gx`` should be a sequence of
        ``nx - 1`` non-negative floats (where ``nx`` is the number of cells),
        and ``gy`` should be ``None``.
        For 2d simulations ``gx`` should be convertible to a numpy array with
        shape ``(ny, nx - 1)`` and ``gy`` should be convertible to a numpy
        array with shape ``(ny - 1, nx)``.

        Calling ``set_conductance_field`` will delete any conductances
        previously set with :meth:`set_conductance` or :meth:`set_connections`.

        If diffusion is disabled, a call to this method will raise a
        ``RuntimeError``.
        """
        if not self._diffusion_enabled:
            raise RuntimeError(
                'This method is unavailable when diffusion is disabled.')

        # Check the field's size
        gx = np.array(gx, copy=False, dtype=float)
        if len(self._dims) == 1:
            s = self._nx - 1
            if gx.shape != (s, ):
                raise ValueError(
                    'The argument `gx` must have length ' + str(s) + '.')
            if gy is not None:
                raise ValueError(
                    'For 1-d simulations the argument `gy` must be None.')
        else:
            s = (self._ny, self._nx - 1)
            if gx.shape != s:
                raise ValueError(
                    'The argument `gx` must have dimensions ' + str(s) + '.')

            if gy is None:
                raise ValueError(
                    'The argument `gy` must be set for 2-d simulations.')
            gy = np.array(gy, copy=False, dtype=float)
            s = (self._ny - 1, self._nx)
            if gy.shape != s:
                raise ValueError(
                    'The argument `gy` must have dimensions ' + str(s) + '.')

        # Check the field values
        if np.any(gx < 0):
            raise ValueError(
                'The argument `gx` can not contain negative values.')
        if (gy is not None) and np.any(gy < 0):
            raise ValueError(
                'The argument `gy` can not contain negative values.')

        # Store
        self._gx_field = list(gx.reshape((self._nx - 1) * self._ny))
        if gy is None:
            self._gy_field = None
        else:
            self._gy_field = list(gy.reshape(self._nx * (self._ny - 1)))
        self._connections = None

    def set_connections(self, connections):
        """
        Adds a list of connections between cells, allowing the creation of
        arbitrary geometries.

        The ``connections`` should be given as a list of tuples
        ``(cell_1, cell_2, conductance)``.

        Connections are only supported in 1d mode -- even though the
        resulting geometry may represent a shape in an arbitrary number of
        dimensions.

        Calling `set_connections` will override any conductances previously set
        with :meth:`set_conductance` or :meth:`set_conductance_field`.

        If diffusion is disabled, a call to this method will raise a
        ``RuntimeError``.
        """
        if not self._diffusion_enabled:
            raise RuntimeError(
                'This method is unavailable when diffusion is disabled.')
        if len(self._dims) != 1:
            raise RuntimeError('Connections can only be specified in 1d mode.')

        if connections is None:
            raise ValueError(
                'Connection list cannot be None. To unset connections, call '
                ' set_conductance() with the new conductance value.')
        conns = []
        doubles = set()
        for x in connections:
            try:
                i, j, c = x
            except Exception:
                raise ValueError(
                    'Connections must be a list of 3-tuples'
                    ' (cell_index_1, cell_index_2, conductance).')

            # Check indices
            i, j = int(i), int(j)
            if i == j or i < 0 or j < 0 or i >= self._nx or j >= self._nx:
                raise ValueError(
                    'Invalid connection: (' + str(i) + ', ' + str(j) + ', '
                    + str(c) + ')')

            # Order indices, and detect doubles
            i, j = (i, j) if i < j else (j, i)
            if (i, j) in doubles:
                raise ValueError(
                    'Duplicate connection: (' + str(i) + ', ' + str(j) + ', '
                    + str(c) + ')')
            doubles.add((i, j))

            # Check conductance
            c = float(c)
            if c < 0:
                raise ValueError('Invalid conductance: ' + str(c))

            # Store connection
            conns.append((i, j, c))
        del doubles
        self._connections = conns
        self._gx_field = self._gy_field = None

    def set_constant(self, var, value):
        """
        Changes a model constant. Only literal constants (constants not
        dependent on any other variable) can be changed.

        The constant ``var`` can be given as a :class:`Variable` or a string
        containing a variable qname. The ``value`` should be given as a float.

        Note that any scalar fields set for the same variable will overwrite
        this value without warning.
        """
        value = float(value)
        if isinstance(var, myokit.Variable):
            var = var.qname()
        var = self._model.get(var)
        if not var.is_literal():
            raise ValueError(
                'The given variable <' + var.qname() + '> is not a literal.')
        # Update value in internal model (will update its defined value when
        # the kernel is generated before the next run).
        self._model.set_value(var.qname(), value)

    def set_default_state(self, state, x=None, y=None):
        """
        Changes this simulation's default state.

        This can be used in three different ways:

        1. When called with an argument ``state`` of size ``n_states`` and
           ``x=None`` the given state will be set as the new default state of
           all cells in the simulation.
        2. Called with an argument ``state`` of size ``n_states`` and
           ``x, y`` equal to a valid cell index, this method will update only
           the selected cell's default state.
        3. Finally, when called with a ``state`` of size
           ``n_states * n_total_cells``, the method will treat ``state`` as a
           concatenation of default state vectors for each cell. For 2d
           simulations, the list should be indexed so that x changes first (the
           first ``nx`` entries have ``y = 0``, the second ``nx`` entries have
           ``y = 1`` and so on).

        """
        self._default_state = self._set_state(state, x, y, self._default_state)

    def set_field(self, var, values):
        """
        Can be used to replace a model constant with a scalar field.

        The argument ``var`` must specify a variable from the simulation's
        model. The field itself is given as ``values``, which must have the
        dimensions ``(ny, nx)``. Multiple fields can be added, depending on the
        memory available on the device. If a field is added for a variable
        already associated with a field, the old data will be overwritten.

        With diffusion currents enabled, this method can let you simulate
        heterogeneous tissue properties. With diffusion disabled, it can be
        used to investigate the effects of changing a parameter through the
        parallel simulation of several cells.
        """
        # Check variable
        if isinstance(var, myokit.Variable):
            var = var.qname()
        var = self._model.get(var)
        if var.is_bound():
            raise ValueError('Bound values cannot be replaced by fields.')
        if not var.is_constant():
            raise ValueError('Only constants can be used for fields.')
        # Check values
        values = np.array(values, copy=False, dtype=float)
        if len(self._dims) == 1:
            if values.shape != (self._nx, ):
                raise ValueError(
                    'The argument `values` must have length ' + str(self._nx)
                    + '.')
        else:
            shape = (self._ny, self._nx)
            if values.shape != shape:
                raise ValueError(
                    'The argument `values` must have dimensions ' + str(shape)
                    + '.')
        # Add field
        self._fields[var] = list(values.reshape(self._nx * self._ny))

    def set_paced_cells(self, nx=5, ny=5, x=0, y=0):
        """
        Sets the number of cells that will receive a stimulus from the pacing
        protocol. For 1d simulations, the values ``ny`` and ``y`` will be
        ignored.

        This method can only define rectangular pacing areas. To select an
        arbitrary set of cells, use :meth:`set_paced_cell_list`.

        Arguments:

        ``nx``
            The number of cells/nodes in the x-direction. If a negative number
            of cells is set the cells left of the offset (``x``) are
            stimulated.
        ``ny``
            The number of cells/nodes in the y-direction. If a negative number
            of cells is set the cells left of the offset (``x``) are
            stimulated.
        ``x``
            The offset of the pacing rectangle in the x-direction. If a
            negative offset is given the offset is calculated from right to
            left.
        ``y``
            The offset of the pacing rectangle in the y-direction. If a
            negative offset is given the offset is calculated from bottom to
            top.

        If diffusion is disabled, all cells are paced and a call to this method
        will raise a ``RuntimeError``.
        """
        if not self._diffusion_enabled:
            raise RuntimeError(
                'This method is unavailable when diffusion is disabled.')

        # Check nx and x. Allow cell selections outside of the boders!
        nx = int(nx)
        x = int(x)
        if nx < 0:
            nx = -nx
            x -= nx
        if x < 0:
            x += self._nx

        # Check dimensions
        if len(self._dims) == 1:
            # Use default for y
            ny = 1
            y = 0
        else:
            # Check ny and y
            ny = int(ny)
            if ny < 0:
                ny = -ny
                y -= ny
            if y < 0:
                y += self._ny

        # Set tuple of paced cells
        self._paced_cells = (nx, ny, x, y)

    def set_paced_cell_list(self, cells):
        """
        Selects the cells to be paced using a list of cell indices. In 1d
        simulations a cell index is an integer ``i``, in 2d simulations cell
        indices are specified as tuples ``(i, j)``.

        For large numbers of cells, this method becomes very inefficient. In
        these cases it may be better to use a rectangular pacing area set using
        :meth:`set_paced_cells`.

        If diffusion is disabled, all cells are paced and a call to this method
        will raise a ``RuntimeError``.
        """
        if not self._diffusion_enabled:
            raise RuntimeError(
                'This method is unavailable when diffusion is disabled.')

        paced_cells = []
        if len(self._dims) == 1:
            for cell in cells:
                cell = int(cell)
                if cell < 0 or cell >= self._nx:
                    raise IndexError(
                        'Cell index out of range: ' + str(cell) + '.')
                paced_cells.append(cell)
        else:
            for i, j in cells:
                i, j = int(i), int(j)
                if i < 0 or j < 0 or i >= self._nx or j >= self._ny:
                    raise IndexError(
                        'Cell index out of range: (' + str(i) + ', ' + str(j)
                        + ').')
                paced_cells.append(i + j * self._nx)
        # Set list of paced cells
        self._paced_cells = paced_cells

    def set_protocol(self, protocol=None):
        """
        Changes the pacing protocol used by this simulation.
        """
        if protocol is None:
            self._protocol = None
        else:
            self._protocol = protocol.clone()

    def _set_state(self, state, x, y, update):
        """
        Handles set_state and set_default_state.
        """
        n = len(state)

        # Set full simulation state
        if n == self._nstate * self._ntotal:
            if x is not None:
                raise ValueError(
                    'A state for all cells was passed in, but argument x'
                    ' was not None.')
            if y is not None:
                raise ValueError(
                    'A state for all cells was passed in, but argument y'
                    ' was not None.')
            return list(state)
        elif n != self._nstate:
            raise ValueError(
                'Given state must have the same size as a single'
                ' cell state or a full simulation state')

        # Set all cells to same state
        if x is None:
            # State might not be a list, at this point, so cast
            return list(state) * self._ntotal

        # Set specific cell state
        x = int(x or 0)
        if x < 0 or x >= self._nx:
            raise IndexError('Given x-index out of range.')
        y = int(y or 0)
        if len(self._dims) == 2:
            if y < 0 or y >= self._ny:
                raise IndexError('Given y-index out of range.')
            x += y * self._nx
        elif y != 0:
            raise ValueError('Y-index given for a 1-dimensional simulation.')
        offset = x * self._nstate
        update[offset:offset + self._nstate] = state
        return update

    def set_state(self, state, x=None, y=None):
        """
        Changes the state of this simulation's model.

        This can be used in three different ways:

        1. When called with an argument ``state`` of size ``n_states`` and
           ``x=None`` the given state will be set as the new state of all
           cells in the simulation.
        2. Called with an argument ``state`` of size ``n_states`` and
           ``x, y`` equal to a valid cell index, this method will update only
           the selected cell's state.
        3. Finally, when called with a ``state`` of size
           ``n_states * n_total_cells``, the method will treat ``state`` as a
           concatenation of state vectors for each cell. For 2d simulations,
           the list should be indexed so that x changes first (the first ``nx``
           entries have ``y = 0``, the second ``nx`` entries have ``y = 1`` and
           so on).

        """
        self._state = self._set_state(state, x, y, self._state)

    def set_step_size(self, step_size=0.005):
        """
        Sets the step size used in the forward Euler solving routine.
        """
        step_size = float(step_size)
        if step_size <= 0:
            raise ValueError('Step size must be greater than zero.')
        self._step_size = step_size

    def set_time(self, time=0):
        """ Sets the current simulation time. """
        self._time = float(time)

    def shape(self):
        """
        Returns the shape of this simulation's grid of cells as a tuple
        ``(ny, nx)`` for 2d simulations, or a single value ``nx`` for 1d
        simulations.
        """
        if len(self._dims) == 2:
            return (self._ny, self._nx)
        return self._nx

    def state(self, x=None, y=None):
        """
        Returns the current simulation state as a list of
        ``len(state) * n_total_cells`` floating point values, where
        ``n_total_cells`` is the total number of cells.

        If the optional arguments ``x`` and ``y`` specify a valid cell index a
        single cell's state is returned. For example ``state(4)`` can be
        used with a 1d simulation, while ``state(4, 2)`` is a valid index in
        the 2d case.

        For 2d simulations, the list is indexed so that x changes first (the
        first ``nx`` entries have ``y = 0``, the second ``nx`` entries have
        ``y = 1`` and so on).
        """
        if x is None:
            return list(self._state)
        else:
            x = int(x or 0)
            if x < 0 or x >= self._nx:
                raise IndexError('Given x-index out of range.')
            y = int(y or 0)
            if len(self._dims) == 2:
                if y < 0 or y >= self._ny:
                    raise IndexError('Given y-index out of range.')
                x += y * self._nx
            elif y != 0:
                raise ValueError(
                    'Y-coordinate specified for 1-dimensional simulation.')

            return self._state[x * self._nstate:(x + 1) * self._nstate]

    def step_size(self):
        """
        Returns the current step size.
        """
        return self._step_size

    def time(self):
        """
        Returns the current simulation time.
        """
        return self._time




#
# Provides extra libraries on top of Myokit's core.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
"""
Provides functionality built on Myokit's core.
"""




#
# Provides an export to OpenCL
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from ._exporter import OpenCLExporter, OpenCLRLExporter
from ._ewriter import OpenCLExpressionWriter
from myokit.formats import ansic


# Importers
# Exporters
_exporters = {
    'opencl': OpenCLExporter,
    'opencl-rl': OpenCLRLExporter,
}


def exporters():
    """
    Returns a dict of all exporters available in this module.
    """
    return dict(_exporters)


# Expression writers
_ewriters = {
    'opencl': OpenCLExpressionWriter,
}


def ewriters():
    """
    Returns a dict of all expression writers available in this module.
    """
    return dict(_ewriters)


# Keywords
keywords = list(ansic.keywords)
keywords.extend([
    # Data types
    'bool',
    'char',
    'uchar',
    'short',
    'ushort',
    'int',
    'uint',
    'long',
    'ulong',
    'float',
    'half',
    'size_t',
    'ptrdiff_t',
    'intptr_t',
    'uintptr_t',
    'void',
    'unsigned',
    # Vector data types
    'char2', 'char4', 'char8', 'char16',
    'uchar2', 'uchar4', 'uchar8', 'uchar16',
    'short2', 'short4', 'short8', 'short16',
    'ushort2', 'ushort4', 'ushort8', 'ushort16',
    'int2', 'int4', 'int8', 'int16',
    'uint2', 'uint4', 'uint8', 'uint16',
    'long2', 'long4', 'long8', 'long16',
    'ulong2', 'ulong4', 'ulong8', 'ulong16',
    'float2', 'float4', 'float8', 'float16',
    # Other data types
    'image2d_t',
    'image3d_t',
    'sampler_t',
    'event_t',
    # Reserved data types
    'bool2', 'bool4', 'bool8', 'bool16',
    'double',
    'double2', 'double4', 'double8', 'double16',
    'half2', 'half4', 'half8', 'half16',
    'quad',
    'quad2', 'quad4', 'quad8', 'quad16',
    'complex half',
    'complex half2', 'complex half4', 'complex half8', 'complex half16',
    'imaginary half',
    'imaginary half2', 'imaginary half4', 'imaginary half8',
    'imaginary half16',
    'complex float',
    'complex float2', 'complex float4', 'complex float8', 'complex float16',
    'imaginary float',
    'imaginary float2', 'imaginary float4', 'imaginary float8',
    'imaginary float16',
    'complex double',
    'complex double2', 'complex double4', 'complex double8',
    'complex double16',
    'imaginary double',
    'imaginary double2', 'imaginary double4', 'imaginary double8',
    'imaginary double16',
    'complex quad',
    'complex quad2', 'complex quad4', 'complex quad8', 'complex quad16',
    'imaginary quad',
    'imaginary quad2', 'imaginary quad4', 'imaginary quad8',
    'imaginary quad16',
    'float2x2', 'float2x4', 'float2x8', 'float2x16',
    'float4x2', 'float4x4', 'float4x8', 'float4x16',
    'float8x2', 'float8x4', 'float8x8', 'float8x16',
    'float16x2', 'float16x4', 'float16x8', 'float16x16',
    'double2x2', 'double2x4', 'double2x8', 'double2x16',
    'double4x2', 'double4x4', 'double4x8', 'double4x16',
    'double8x2', 'double8x4', 'double8x8', 'double8x16',
    'double16x2', 'double16x4', 'double16x8', 'double16x16',
    'long double',
    'long double2', 'long double4', 'long double8', 'long double16',
    'long long',
    'long long2', 'long long4', 'long long8', 'long long16',
    'unsigned long long',
    'ulong long',
    'ulong long2', 'ulong long4', 'ulong long8', 'ulong long16',
    # Address space qualifiers
    '__global',
    '__local',
    '__constant',
    '__private',
    # Image access qualifiers
    '__read_only',
    '__write_only',
    # Function qualifiers
    '__kernel',
    '__attribute__',
    # ^ These don't really need to be here, since variables names never start
    # on an underscore in myokit.
    # Work-item functions
    'get_work_dim',
    'get_global_size',
    'get_global_id',
    'get_local_size',
    'get_local_id',
    'get_num_groups',
    'get_group_id',
    # Math functions
    'acosh',
    'acospi',
    'asin',
    'asinh',
    'asinpi',
    'atan',
    'atan2',
    'atanh',
    'atanpi',
    'atan2pi',
    'cbrt',
    'ceil',
    'copysign',
    'cos',
    'cosh',
    'cospi',
    'erfc',
    'erf',
    'exp',
    'exp2',
    'exp10',
    'expm1',
    'fabs',
    'fdim',
    'floor',
    'fma',
    'fmax',
    'fmin',
    'fmod',
    'fract',
    'frexp',
    'hypot',
    'ilogb',
    'ldexp',
    'lgamma',
    'lgamma_r',
    'log',
    'log2',
    'log10',
    'log1p',
    'logb',
    'mad',
    'modf',
    'nan',
    'nextafter',
    'pow',
    'pown',
    'powr',
    'remainder',
    'remquo',
    'rint',
    'rootn',
    'round',
    'rsqrt',
    'sin',
    'sincos',
    'sinh',
    'sinpi',
    'sqrt',
    'tan',
    'tanh',
    'tanpi',
    'tgamma',
    'trunc',
    # Special math functions
    'half_cos',
    'half_divide',
    'half_exp',
    'half_exp2',
    'half_exp10',
    'half_log',
    'half_log2',
    'half_log10',
    'half_powr',
    'half_recip',
    'half_rsqrt',
    'half_sin',
    'half_sqrt',
    'half_tan',
    'native_cos',
    'native_divide',
    'native_exp',
    'native_exp2',
    'native_exp10',
    'native_log',
    'native_log2',
    'native_log10',
    'native_powr',
    'native_recip',
    'native_rsqrt',
    'native_sin',
    'native_sqrt',
    'native_tan',
    # Math constants
    'MAXFLOAT',
    'HUGE_VALF',
    'INFINITY',
    'NAN',
    # Math macros
    'FLT_DIG',
    'FLT_MANT_DIG',
    'FLT_MAX_10_EXP',
    'FLT_MAX_EXP',
    'FLT_MIN_10_EXP',
    'FLT_MIN_EXP',
    'FLT_RADIX',
    'FLT_MAX',
    'FLT_MIN',
    'FLT_EPSILON',
    # Integer functions
    'abs',
    'abs_diff',
    'add_sat',
    'hadd',
    'rhadd',
    'clz',
    'mad_hi',
    'mad_sat',
    'max',
    'min',
    'mul_hi',
    'rotate',
    'sub_sat',
    'upsample',
    'mad24',
    'mul24',
    # Integer macros
    'CHAR_BIT',
    'CHAR_MAX',
    'CHAR_MIN',
    'INT_MAX',
    'INT_MIN',
    'LONG_MAX',
    'LONG_MIN',
    'SCHAR_MAX',
    'SCHAR_MIN',
    'SHRT_MAX',
    'SHRT_MIN',
    'UCHAR_MAX',
    'USHRT_MAX',
    'UINT_MAX',
    'ULONG_MAX',
    # Common functions
    'clamp',
    'degrees',
    'max',
    'min',
    'mix',
    'radians',
    'step',
    'smoothstep',
    'sign',
    # Geometric functions
    'cross',
    'dot',
    'distance',
    'length',
    'normalize',
    'fast_distance',
    'fast_length',
    'fast_normalize',
    # Relational functions
    'isequal',
    'isnotequal',
    'isgreater',
    'isgreaterequal',
    'isless',
    'islessequal',
    'islessgreater',
    'isfinite',
    'isinf',
    'isnan',
    'isnormal',
    'isordered',
    'isunordered',
    'signbit',
    'any',
    'all',
    'bitselect',
    'select',
    # Vector data
    'vload2', 'vload4', 'vload8', 'vload16',
    'vstore2', 'vstore4', 'vstore8', 'vstore16',
    'vload_half',
    'vload_half2', 'vload_half4', 'vload_half8', 'vload_half16',
    'vstore_half',
    'vstore_half2', 'vstore_half4', 'vstore_half8', 'vstore_half16',
    # Other
    'union',
    'atomic_cmpxchg',
    #TODO
    # There's still more :)
])
